"""
Host/controller interface adaptation layer.
"""
import logging
from time import sleep
from queue import Queue, Empty
from struct import unpack
from threading import Lock

# Scapy layers for HCI
from scapy.layers.bluetooth import BluetoothSocketError, BluetoothUserSocket, \
    HCI_Hdr, HCI_Command_Hdr, HCI_Cmd_Reset, HCI_Cmd_Set_Event_Filter, \
    HCI_Cmd_Set_Event_Mask, HCI_Cmd_Write_LE_Host_Support, \
    HCI_Cmd_Read_BD_Addr, HCI_Cmd_Complete_Read_BD_Addr, HCI_Cmd_LE_Set_Scan_Enable, \
    HCI_Cmd_LE_Set_Scan_Parameters, HCI_Cmd_LE_Create_Connection, HCI_Cmd_Disconnect, \
    HCI_Cmd_LE_Set_Advertise_Enable, HCI_Cmd_LE_Set_Advertising_Data, \
    HCI_Event_Disconnection_Complete, HCI_Cmd_LE_Set_Scan_Response_Data, \
    HCI_Cmd_LE_Set_Random_Address, HCI_Cmd_LE_Long_Term_Key_Request_Reply, \
    HCI_Cmd_LE_Enable_Encryption, HCI_Cmd_LE_Set_Advertising_Parameters, \
    HCI_Cmd_LE_Read_Buffer_Size_V1, HCI_Cmd_Read_Local_Name, HCI_Cmd_Complete_Read_Local_Name, \
    HCI_Cmd_Complete_Read_Local_Version_Information, HCI_Cmd_Read_Local_Version_Information, \
    HCI_Cmd_Write_Connect_Accept_Timeout, HCI_Cmd_LE_Read_Local_Supported_Features, \
    HCI_Cmd_LE_Read_Filter_Accept_List_Size, HCI_Cmd_LE_Clear_Filter_Accept_List, \
    EIR_Hdr, HCI_Cmd_LE_Create_Connection_Cancel

from whad.scapy.layers.bluetooth import HCI_Cmd_LE_Complete_Read_Buffer_Size, \
    HCI_Cmd_Read_Buffer_Size, HCI_Cmd_Complete_Read_Buffer_Size, HCI_Cmd_LE_Set_Event_Mask, \
    HCI_Cmd_Read_Local_Supported_Commands, HCI_Cmd_Complete_Supported_Commands, \
    HCI_Cmd_Read_Local_Supported_Features, HCI_Cmd_Complete_Supported_Features, \
    HCI_Cmd_LE_Complete_Read_Filter_Accept_List_Size, HCI_Cmd_LE_Complete_Supported_Features, \
    HCI_Cmd_LE_Write_Suggested_Default_Data_Length, HCI_Cmd_LE_Read_Suggested_Default_Data_Length, \
    HCI_Cmd_LE_Complete_Suggested_Default_Data_Length, HCI_Cmd_Write_Simple_Pairing_Mode, \
    HCI_Cmd_Write_Default_Link_Policy_Settings, HCI_Cmd_LE_Read_Advertising_Physical_Channel_Tx_Power, \
    HCI_Cmd_Complete_LE_Advertising_Tx_Power_Level, HCI_Cmd_Write_Class_Of_Device

# Whad custom layers
from whad.scapy.layers.hci import HCI_VERSIONS, BT_MANUFACTURERS, \
    HCI_Cmd_LE_Read_Supported_States, \
    HCI_Cmd_Complete_LE_Read_Supported_States, HCI_Cmd_CSR_Write_BD_Address, HCI_Cmd_CSR_Reset, \
    HCI_Cmd_TI_Write_BD_Address, HCI_Cmd_BCM_Write_BD_Address, HCI_Cmd_Zeevo_Write_BD_Address, \
    HCI_Cmd_Ericsson_Write_BD_Address, HCI_Cmd_ST_Write_BD_Address, \
    HCI_Cmd_LE_Set_Host_Channel_Classification

# Whad
from whad.exceptions import WhadDeviceNotFound, WhadDeviceNotReady, WhadDeviceAccessDenied, \
    WhadDeviceUnsupportedOperation, WhadDeviceError

# Whad hub
from whad.hub.discovery import Domain
from whad.hub.generic.cmdresult import CommandResult
from whad.hub.discovery import Capability
from whad.hub.ble import Direction as BleDirection, Commands, AddressType, BDAddress

from ..device import VirtualDevice
from .converter import HCIConverter
from .hciconfig import HCIConfig
from .constants import LE_STATES, ADDRESS_MODIFICATION_VENDORS, HCIInternalState, \
    HCIConnectionState

logger = logging.getLogger(__name__)

def get_hci(index):
    '''
    Returns an HCI socket based on adapter index.
    '''
    try:
        logger.debug("Creating bluetooth socket ...")
        socket = BluetoothUserSocket(index)
        logger.debug("Bluetooth socket successfully created.")
        return socket
    except BluetoothSocketError:
        logger.debug("An error occurred while creating bluetooth socket")
        try:
            logger.debug("Shutting down HCI interface #%d", index)
            HCIConfig.down(index)
            logger.debug("HCI interface %d shut down, creating Bluetooth socket ...", index)
            socket = BluetoothUserSocket(index)
            logger.debug("Bluetooth socket successfully created.")
            return socket
        except BluetoothSocketError as err:
            logger.debug(err)
            logger.debug("Cannot create Bluetooth socket !")
            return None
        except PermissionError as perm_err:
            logger.debug("WHAD device hci%d cannot be accessed.", index)
            raise WhadDeviceAccessDenied(f"hci{index}") from perm_err
    except PermissionError as perm_err:
        logger.debug("WHAD device hci%d cannot be accessed.", index)
        raise WhadDeviceAccessDenied(f"hci{index}") from perm_err

def compute_max_time(length: int, datarate: int) -> int:
    """Compute the maximum transmission time for a given PDU length and
    datarate.

    Prefix is 80 bits (header + CRC), suffix is some kind of security margin.
    """
    return int(((80 + length*8 + 32)/datarate)*1000000.0)

class HCIUnsupportedCommand(Exception):
    """Raised when an HCI command requirement is not met by hardware.
    """
    def __init__(self, command):
        super().__init__()
        self.command = command

    def __repr__(self):
        return f"HCIUnsupportedCommand(cmd='{self.command}')"

class HCIUnsupportedFeature(Exception):
    """Raised when an HCI feature requirement is not met by hardware.
    """
    def __init__(self, feature):
        super().__init__()
        self.__feature = feature

    def __repr__(self):
        return f"HCIUnsupportedFeature(feature='{self.__feature}')"

class HCIUnsupportedLEFeature(Exception):
    """Raised when an HCI LE feature requirement is not met by hardware.
    """
    def __init__(self, feature):
        super().__init__()
        self.__feature = feature

    def __repr__(self):
        return f"HCIUnsupportedLEFeature(feature='{self.__feature}')"

class req_cmd:
    """HCI Decorator to handle command requirements.

    Normally, we query an HCI interface to retrieve the list of its supported
    commands, following the recommended initialization procedure
    (Vol6, part D, section 1).

    We need to check that all the required commands are supported by the target
    hardware before starting a specific procedure, and this decorator provides
    a way to declare one or more required commands for a decorated method of
    HciIface, and blocks its execution if at least one of them is not provided
    by the target hardware.
    """

    def __init__(self, *args):
        """Save any string argument as a required HCI command.
        """
        self.__requires = []
        for arg in args:
            if isinstance(arg, str):
                self.__requires.append(arg)

    def __call__(self, method):
        """Called to decorate the actual method.
        """
        requirements = self.__requires
        def _wrap(self, *args, **kwargs):
            # check our requirements are met
            for command in requirements:
                if not self.is_cmd_supported(command):
                    raise HCIUnsupportedCommand(command)

            # If all requirements are met, forward
            return method(self, *args, **kwargs)
        return _wrap

class req_feature:
    """HCI decorator to handle command feature requirement.
    """
    def __init__(self, *args):
        self.__requires = []
        for arg in args:
            if isinstance(arg, str):
                self.__requires.append(arg)

    def __call__(self, method):
        """Called to decorate the actual method.
        """
        requirements = self.__requires
        def _wrap(self, *args, **kwargs):
            # check our requirements are met
            for feature in requirements:
                if not self.is_feature_supported(feature):
                    raise HCIUnsupportedFeature(feature)

            # If all requirements are met, forward
            return method(self, *args, **kwargs)
        return _wrap 

class le_only(req_feature):
    """Requires a LE-enabled controller
    """
    def __init__(self, *args):
        super().__init__("le_supported_controller")
        self.__requires = []
        for arg in args:
            if isinstance(arg, str):
                self.__requires.append(arg)

    def __call__(self, method):
        requirements = self.__requires
        def _wrap(self, *args, **kwargs):
            # check our requirements are met
            for feature in requirements:
                if not self.is_le_feature_supported(feature):
                    raise HCIUnsupportedLEFeature(feature)

            # If all requirements are met, forward
            return method(self, *args, **kwargs)

        # Wrap with LE-enabled controller check (tested first)
        return super().__init__(_wrap)

class Hci(VirtualDevice):
    """Host/controller interface virtual device implementation.
    """

    INTERFACE_NAME = "hci"

    PHY_1M = 1000000
    PHY_2M = 2000000

    @classmethod
    def list(cls):
        '''
        Returns a list of available HCI devices.
        '''
        available_devices = {}
        devices_ids = HCIConfig.list()
        for device_id in devices_ids:
            available_devices[device_id] = Hci(index=device_id)

        return available_devices

    def __init__(self, index):
        super().__init__(index=index)
        self.__converter = HCIConverter(self)
        self.__index = index
        self.__lock = Lock()
        self.__socket = None
        self.__internal_state = HCIInternalState.NONE
        self.__conn_state = HCIConnectionState.DISCONNECTED
        self.__opened = False
        self.__hci_responses = Queue()
        self._dev_capabilities = None
        self._local_name = None
        self._advertising = False
        self._bd_address = None
        self._bd_address_type = AddressType.PUBLIC
        self._fw_version = None
        self._fw_url = None
        self._fw_author = None
        self._dev_id = None
        self._manufacturer = None
        self._cached_scan_data = None
        self._cached_scan_response_data = None
        self.__timeout = 1.0
        self._connected = False
        self.__closing = False
        self.__local_supp_cmds = None

        # Data PDU Length management
        self.__datarate = Hci.PHY_1M
        self.__conn_max_tx_octets = 27
        self.__conn_max_tx_time = 0x148
        self.__conn_max_tx_time_uncoded = 328
        self.__conn_max_tx_time_coded = 2704
        self.__conn_max_rx_octets = 27
        self.__conn_max_rx_time = 0x148
        self.__supported_max_tx_octets = 27
        self.__supported_max_tx_time = 0x148
        self.__supported_max_rx_octets = 27
        self.__supported_max_rx_time = 0x148
        self._active_handles = []

        # Classic Features
        self.__features = None

        # LE Features
        self.__le_features = None

        # LE Filter Accept
        self.__fa_size = 0
        self.__fa_entries = []

    @property
    def identifier(self):
        '''
        Returns the identifier of the device (e.g., index).
        '''
        return "hci" + str(self.__index)


    def open(self):
        """
        Open device.
        """
        if not self.__opened:
            # Open a Bluetooth user socket
            self.__socket = get_hci(self.__index)
            if self.__socket is None:
                # An error occurred, raising error
                logger.error("Whad device is not ready.")
                raise WhadDeviceNotReady()

            # Flush HCI interface
            logger.debug("[%s] Flushing HCI interface ...", self.interface)
            self.__socket.flush()

            # Mark as opened
            self.__opened = True

            # Ask parent class to run a background I/O thread
            super().open()

            # Initialize this HCI interface
            if not self._initialize():
                raise WhadDeviceNotReady()

    def close(self):
        """
        Close current device.
        """
        # Avoid recursion that may occur due to super().close()
        if self.__closing:
            return

        # Marking device as in closing process
        self.__closing = True

        if self.__conn_state == HCIConnectionState.ESTABLISHED:
            # Disconnect if necessary
            for handle in self._active_handles:
                self._disconnect(handle)
        elif self.__conn_state == HCIConnectionState.INITIATING:
            # Cancel current connection if still trying to connect
            self.cancel_connection()

        # Stop advertising if enabled
        if self._advertising:
            self._set_advertising_mode(False)

        # Ask parent class to stop I/O thread
        logger.debug("Stopping background IO threads ...")
        super().close()

        # Close underlying device.
        if self.__socket is not None:
            logger.debug("Closing Bluetooth socket ...")
            self.__socket.close()
            del self.__socket
            self.__socket = None
        self.__opened = False

        # Closing process done.
        self.__closing = False


    def write(self, payload):
        """
        Writes data to the device. It relies on select() in order to make sure
        we are allowed to write to the device and wait without eating too much CPU
        if the device is not ready to be written to.

        :param bytes data: Data to write
        :return: number of bytes written to the device
        """
        if not self.__opened:
            raise WhadDeviceNotReady()
        self.__socket.send(payload)

    def read(self):
        """
        Fetches data from the device, if there is any data to read. We call select()
        to make sure data is waiting to be read before reading it.
        """
        if not self.__opened:
            raise WhadDeviceNotReady()
        try:
            # We assume here that we can determine if there is something to read
            # by calling readable without locking our socket
            if self.__socket is not None and self.__socket.readable(0.1):
                # Lock our socket to catch the awaiting event
                self.__lock.acquire()
                event = self.__socket.recv()
                self.__lock.release()
                if event.type == 0x4 and event.code in (0xe, 0xf, 0x13):
                    self.__hci_responses.put(event)
                else:
                    messages = self.__converter.process_event(event)
                    if messages is not None:
                        for message in messages:
                            self._send_whad_message(message)

                    # If the connection is stopped and peripheral mode is started,
                    # automatically re-enable advertising based on cached data
                    if HCI_Event_Disconnection_Complete in event:
                        logger.debug("[hci] Disconnection complete")
                    if HCI_Event_Disconnection_Complete in event and \
                            self.__internal_state == HCIInternalState.PERIPHERAL:

                        # If advertising was not enabled, skip
                        if not self._advertising:
                            return

                        if not self._read_advertising_physical_channel_tx_power(from_queue=False):
                            logger.error("[%s] Cannot read advertising physical channel Tx power", self.interface)

                        # if data are cached, configure them
                        if self._cached_scan_data is not None:
                            # We can't wait for response because we are in the
                            # reception loop context
                            success = self._set_advertising_data(self._cached_scan_data,
                                                                 from_queue=False)

                            # Raise an error if we cannot set the advertising data.
                            if not success:
                                logger.error("[hci] cannot set advertising data!")
                                raise WhadDeviceNotReady()

                        if self._cached_scan_response_data is not None:
                            success = self._set_scan_response_data(
                                self._cached_scan_response_data, from_queue=False
                            )

                            # Raise an error if we cannot set the scan response data.
                            if not success:
                                logger.error("[hci] cannot set scan response data!")
                                raise WhadDeviceNotReady()

                        # We need to artificially disable advertising indicator
                        # to prevent cached operation
                        self._advertising = False
                        self._set_advertising_mode(True, from_queue=False)


        except (BrokenPipeError, OSError) as err:
            print(err)
            logger.error("Error, waiting...")
            sleep(1)

    def _wait_response(self, timeout=None):
        response = None
        try:
            response = self.__hci_responses.get(block=True, timeout=timeout)
        except Empty:
            logger.debug("[hci] device did not respond")
        return response

    def _write_packet(self, packet):
        """
        Writes an HCI packet.
        """
        logger.debug("[hci] sending packet ...")

        # We claim access to our socket by acquiring its lock
        self.__lock.acquire()

        # And we send our HCI ACL packet
        self.__socket.send(packet)

        # We will wait for a response packet in our HCI RX queue, so we
        # can release our socket to let other threads handle it.
        self.__lock.release()

        # Wait for response
        logger.debug("[hci] waiting for response (timeout: %s)...", self.__timeout)
        response = self._wait_response(timeout=self.__timeout)
        if response is None:
            logger.debug("[hci][%s] timeout reached when sending packet", self.interface)
            return False

        logger.debug("[hci] response code: 0x%04x", response.code)
        while response.code != 0x13:
            logger.debug("[hci] waiting for response ...")
            response = self._wait_response(timeout=self.__timeout)
            if response is None:
                logger.debug("[hci][%s] wait_response returned None !", self.interface)
                return False
            logger.debug("[hci] response code: 0x%04x", response.code)
        return response.num_handles == 1 and response.num_completed_packets_list[0] == 1

    def _write_command(self, command, from_queue=True):
        """
        Writes an HCI command and returns the response.
        """
        # Prepare HCI command
        hci_command = HCI_Hdr()/HCI_Command_Hdr()/command

        # Acquire lock on our socket
        self.__lock.acquire()

        # Send prepared HCI command
        logger.debug("[%s][write_command] Sending HCI command to user socket ...", self.interface)
        self.__socket.send(hci_command)
        logger.debug("[%s][write_command] Command sent.", self.interface)

        # If we are expecting to receive events from our reception queue, we
        # need to release the socket lock and wait for our read() method to
        # catch an answer. This is usually done when an HCI command is initiated
        # by the user application or the protocol stack in use.
        if from_queue:
            # We release our socket lock
            self.__lock.release()

            # And we wait for a response to be sent to our reception queue
            logger.debug("[%s][write_command] Waiting for response ...", self.interface)
            response = self._wait_response()
            while response.opcode != hci_command[HCI_Command_Hdr].opcode:
                logger.debug("[%s][write_command] Received response with opcode %d", self.interface, response.opcode)
                response = self._wait_response()
            logger.debug("[%s][write_command] Response received.", self.interface)

            if response is not None:
                logger.debug("[%s] HCI write command returned status %d",
                            self.interface, response.status)
        else:
            # In the other case, _write_command() is directly called from the
            # read() method and therefore we need to keep our socket locked
            # in order to read from it. Any event received that is not an
            # expected HCI_Command_Complete message is sent to the HCI RX
            # queue.

            event = self.__socket.recv()
            while not (event.type == 0x4 and event.code == 0xe):
                if event.type == 0x4 and event.code in (0xf, 0x13):
                    self.__hci_responses.put(event)
                event = self.__socket.recv()

            # We got our response: we release our socket lock and set the
            # captured event as the reponse to return to caller.
            self.__lock.release()
            response = event

        return response

    def reset(self):
        """
        self._bd_address = self._read_bd_address()
        self._local_name = self._read_local_name()
        """
        self._initialize()
        self._read_local_name()
        self._fw_version, self._manufacturer = self._read_local_version_information()
        self._fw_author = self._manufacturer
        self._dev_id = self._generate_dev_id()
        self._fw_url = b"<unknown>"
        self._dev_capabilities = self._get_capabilities()

    def _generate_dev_id(self):
        devid = (self._bd_address.value + self._local_name)[:16]
        if len(devid) < 16:
            devid += b"\x00" * (16 - len(devid))
        return devid

    def _get_capabilities(self):
        supported_states = self._read_le_supported_states()

        capabilities = 0
        supported_commands = []
        for state in supported_states:
            _, cap, commands = state
            capabilities = capabilities | cap
            supported_commands += commands

        supported_commands += [Commands.SetBdAddress]

        supported_commands = list(set(supported_commands))
        capabilities = {
            Domain.BtLE : (
                                capabilities | Capability.NoRawData,
                                supported_commands
            )
        }
        return capabilities

    def is_cmd_supported(self, cmd: str) -> bool:
        """Determine if a specific LE command is supported by the HCI interface.

        :param cmd: Command to test
        :type cmd: str
        :return: True if command is supported, False otherwise
        :rtype: bool
        """
        if cmd in self.__local_supp_cmds.supported_commands.names:
            return getattr(self.__local_supp_cmds.supported_commands, cmd)
        return False

    def is_feature_supported(self, feature: str) -> bool:
        """Determine if a specific feature is supported by the HCI interface.

        :param feature: Feature to test
        :type feature: str
        :return: True if feature is supported, False otherwise
        :rtype: bool
        """
        if feature in self.__features.lmp_features.names:
            return getattr(self.__features.lmp_features, feature)
        return False
    
    def is_le_feature_supported(self, feature: str) -> bool:
        """Determine if a specific feature is supported by the HCI interface.

        :param feature: Feature to test
        :type feature: str
        :return: True if feature is supported, False otherwise
        :rtype: bool
        """
        if feature in self.__le_features.le_features.names:
            return getattr(self.__le_features.le_features, feature)
        return False

    def _reset(self):
        """
        Reset HCI device.
        """
        logger.debug("[%s] Resetting interface ...", self.interface)
        response = self._write_command(HCI_Cmd_Reset())
        return response is not None and response.status == 0x0

    @req_cmd("read_buffer_size")
    def _read_buffer_size(self):
        """Read HCI device default buffer size and update max ACL length.
        """
        logger.debug("[%s] Reading HCI ACL buffer size ...", self.interface)
        response = self._write_command(HCI_Cmd_Read_Buffer_Size())

        if response is not None and response.status == 0x0:
            if HCI_Cmd_Complete_Read_Buffer_Size in response:
                if response.acl_pkt_len > 0:
                    # Update HCI MTU
                    logger.debug("[%s] ACL buffer length: %d", self.interface, response.acl_pkt_len)
                    self.__conn_max_tx_octets = response.acl_pkt_len
                    return True

        logger.debug("[%s] Failed reading ACL buffer size !", self.interface)
        return False

    @req_cmd("le_read_buffer_size_v1")
    def _le_read_buffer_size(self):
        """Read HCI device LE buffer size and update max ACL length.
        """
        logger.debug("[%s] Reading HCI LE ACL buffer size v1 ...", self.interface)
        response = self._write_command(HCI_Cmd_LE_Read_Buffer_Size_V1())
        if response is not None and response.status == 0x0:
            if HCI_Cmd_LE_Complete_Read_Buffer_Size in response:
                if response.acl_pkt_len > 0:
                    # Update HCI MTU
                    logger.debug("[%s] LE ACL buffer length: %d", self.interface,
                                 response.acl_pkt_len)
                    self.__conn_max_tx_octets = response.acl_pkt_len
                    return True
                else:
                    logger.debug("[%s] LE ACL buffer is 0, fallback to default ACL buffer")
                    return self._read_buffer_size()

        logger.debug("[%s] Failed reading LE ACL buffer size v1 !", self.interface)
        return False

    def read_local_supported_commands(self):
        """Read local adapter supported commands.
        """
        logger.debug("[%s] Reading HCI local supported commands ...", self.interface)
        response = self._write_command(HCI_Cmd_Read_Local_Supported_Commands())
        if response is not None and response.status == 0x0:
            if HCI_Cmd_Complete_Supported_Commands in response:
                logger.debug("[%s] Local supported commands cached.", self.interface)
                self.__local_supp_cmds = response[HCI_Cmd_Complete_Supported_Commands]
                return True

        logger.debug("[%s] Failed reading supported commands !", self.interface)
        return False

    @req_cmd("read_local_supported_features")
    def read_local_supported_features(self):
        """Query the HCI interface to retrieve its supported features.
        """
        logger.debug("[%s] Reading HCI local supported features ...", self.interface)
        response = self._write_command(HCI_Cmd_Read_Local_Supported_Features())
        if response is not None and HCI_Cmd_Complete_Supported_Features in response:
            logger.debug("[%s] Local supported features cached.", self.interface)
            self.__features = response[HCI_Cmd_Complete_Supported_Features]
            return True

        logger.debug("[%s] Failed reading supported features !", self.interface)
        return False

    @req_cmd("le_read_local_supported_features")
    def read_local_le_supported_features(self):
        """Read local LE supported features
        """
        logger.debug("[%s] Reading HCI LE local supported features ...", self.interface)
        response = self._write_command(HCI_Cmd_LE_Read_Local_Supported_Features())
        if response is not None and HCI_Cmd_LE_Complete_Supported_Features in response:
            logger.debug("[%s] Local LE supported features cached.", self.interface)
            self.__le_features = response[HCI_Cmd_LE_Complete_Supported_Features]
            return True

        logger.debug("[%s] Failed reading LE supported features !", self.interface)
        return False

    @req_cmd("set_event_filter")
    def _set_event_filter(self, filter_type=0):
        """
        Configure HCI device event filter.
        """
        logger.debug("[%s] Setting HCI Event Filter to type %d", self.interface, filter_type)
        response = self._write_command(HCI_Cmd_Set_Event_Filter(type=filter_type))
        return response is not None and response.status == 0x00

    @req_cmd("set_event_mask")
    def _set_event_mask(self, mask=b"\xff\xff\xfb\xff\x07\xf8\xbf\x3d"):
        """
        Configure HCI device event mask.
        """
        logger.debug("[%s] Setting HCI Event Mask to %s", self.interface, mask.hex())
        response = self._write_command(HCI_Cmd_Set_Event_Mask(mask=mask))
        return response is not None and response.status == 0x00

    @req_cmd("le_set_event_mask")
    def _le_set_event_mask(self, mask=b"\x1f\x00\x00\x00\x00\x00\x00\x00"):
        """Configure HCI LE event mask.
        """
        logger.debug("[%s] Setting HCI LE Event Mask to %s", self.interface, mask.hex())
        response = self._write_command(HCI_Cmd_LE_Set_Event_Mask())
        return response is not None and response.status == 0x00

    @req_cmd("write_connection_accept_timeout")
    def _set_connection_accept_timeout(self, timeout=32000):
        """
        Configure HCI device connection accept timeout.
        """
        logger.debug("[%s] Setting connection timeout to %d", self.interface, timeout)
        response = self._write_command(HCI_Cmd_Write_Connect_Accept_Timeout(timeout=timeout))
        return response is not None and response.status == 0x00

    @req_cmd("write_le_host_support")
    def indicates_le_support(self):
        """
        Indicates to HCI Device that the Host supports Low Energy mode.
        """
        logger.debug("[%s] Write LE Host support (simulatenous mode not supported)", self.interface)
        response = self._write_command(HCI_Cmd_Write_LE_Host_Support(supported=1, unused=0))
        return response is not None and response.status == 0x00

    @req_cmd("le_write_suggested_default_data_length",
             "le_read_suggested_default_data_length")
    def configure_data_length(self):
        """Negociate ACL data length.
        """
        # If our controller supports LE Data Length Update, suggest an intermediate
        # size and compute time
        if self.is_le_feature_supported("data_packet_length_extension"):
            suggested_max_tx_octets = 64
            suggested_max_tx_time = compute_max_time(suggested_max_tx_octets, self.__datarate)
        else:
            suggested_max_tx_octets = 27
            suggested_max_tx_time = 0x148

        # Send suggested default data length first
        logger.debug("[%s] HCI Write Suggested Default Data Length (max TX octets:%d, max Tx time: %d)",
                     self.interface, self.__conn_max_tx_octets, self.__conn_max_tx_time)
        response = self._write_command(HCI_Cmd_LE_Write_Suggested_Default_Data_Length(
            max_tx_octets=suggested_max_tx_octets,
            max_tx_time=suggested_max_tx_time
        ))
        if response is not None and response.status == 0x00:
            logger.debug("[%s] Suggested Default Data Length successfully sent.", self.interface)
        else:
            logger.debug("[%s] Failed sending Suggested Default Data Length !", self.interface)

        # Read suggested default data length from controller
        logger.debug("[%s] Reading local Suggested Default Data Length ...", self.interface)
        response = self._write_command(HCI_Cmd_LE_Read_Suggested_Default_Data_Length())
        if response is not None and HCI_Cmd_LE_Complete_Suggested_Default_Data_Length in response:
            # Check if reported data length is greater than 27, if so we need to
            # ensure the controller supports the Data Length Extension and follow
            # this procedure.
            answer = response[HCI_Cmd_LE_Complete_Suggested_Default_Data_Length]
            logger.debug(
                "[%s] Suggested Default Data Length read (max_tx_octets:%d, max_tx_time: %d)",
                self.interface, answer.max_tx_octets, answer.max_tx_time
            )
            return True

        logger.debug("[%s] Failed reading Suggested Default Data Length !", self.interface)
        return False

    def write_simple_pairing_mode(self, enable: bool = True) -> bool:
        """Configure HCI interface to support Simple Pairing mode (or not)"""
        response = self._write_command(HCI_Cmd_Write_Simple_Pairing_Mode(enable=enable))
        return response is not None and response.status == 0x00

    def write_connect_accept_timeout(self, timeout: int = 32000) -> bool:
        """Configure the HCI interface connection timeout for LE and BR/EDR controller"""
        response = self._write_command(HCI_Cmd_Write_Connect_Accept_Timeout(timeout=timeout))
        return response is not None and response.status == 0x00

    def write_default_link_policy_settings(self, policy: int = 0x07) -> bool:
        """Configure the HCI interface default link policy settings for LE and BR/EDR controller"""
        response = self._write_command(HCI_Cmd_Write_Default_Link_Policy_Settings(policy=policy))
        return response is not None and response.status == 0x00

    def write_device_class(self, major_service_class: int = 0x360, major_device_class: int = 0x00,
                           minor_device_class: int = 0x00) -> bool:
        """Configure a default class of device"""
        response = self._write_command(HCI_Cmd_Write_Class_Of_Device(
            major_service_classes=major_service_class,
            major_device_class=major_device_class,
            minor_device_class=minor_device_class
        ))
        return response is not None and response.status == 0x00

    def _initialize(self):
        """
        Initialize HCI Device and returns boolean indicating if it can be used by WHAD.
        """
        logger.debug("[%s] Starting initialization process ...", self.interface)
        success = (
                self._reset() and
                self.read_local_supported_commands() and 
                self.read_local_le_supported_features() and 
                self.write_simple_pairing_mode() and
                self.write_connect_accept_timeout() and
                self.write_device_class() and
                self._set_event_mask(b"\xff\xff\xfb\xff\x07\xf8\xbf\x3d") and
                self._le_set_event_mask(mask=b'\xff\xff\xff\xff\x03') and
                self._le_read_buffer_size() and
                self._read_bd_address() and
                self.indicates_le_support() and
                self.__read_filter_accept_list_size() and
                self.clear_filter_list()
        )
        try:
            success = success and self.configure_data_length()
        except HCIUnsupportedCommand as cmderr:
            logger.debug("[%s] Configuring data length cannot be done, command %s is not supported.",
                         self.interface, cmderr.command)
        logger.debug("[%s] Initialization process result: %s", self.interface,
                     "Success" if success else "Failed")
        return success

    @req_cmd("le_clear_filter_accept_list")
    def clear_filter_list(self):
        """Clear LE filter list.
        """
        logger.debug("[%s] Clearing LE filter accept list ...", self.interface)
        response = self._write_command(HCI_Cmd_LE_Clear_Filter_Accept_List())
        return response is not None and response.status == 0x00

    @req_cmd("le_read_filter_accept_list_size")
    def __read_filter_accept_list_size(self) -> bool:
        """Retrieve LE Filter accept list size from local adapter.
        """
        logger.debug("[%s] Reading LE filter accept list size ...", self.interface)
        response = self._write_command(HCI_Cmd_LE_Read_Filter_Accept_List_Size())
        if response is not None and response.status == 0x00:
            r = response[HCI_Cmd_LE_Complete_Read_Filter_Accept_List_Size]
            self.__fa_size = r.list_size
            logger.debug("[%s] LE filter accept list size: %d slots", self.interface,
                         self.__fa_size)
            return True
        return False

    def get_whitelist_size(self) -> int:
        """Retrieve the LE Device Whitelist size for the current HCI
        interface.

        :return: Number of slots in current LE Whitelist (Filter Accept List)
        :rtype: int
        """
        # Read Filter Accept List size
        self.__read_filter_accept_list_size()
        return self.__fa_size

    @req_cmd("read_bd_addr")
    def _read_bd_address(self):
        """
        Read BD Address used by the HCI device.
        """
        logger.debug("[%s] Reading HCI interface BD address ...", self.interface)
        response = self._write_command(HCI_Cmd_Read_BD_Addr())
        if response.status == 0x00 and HCI_Cmd_Complete_Read_BD_Addr in response:
            self._bd_address = BDAddress(response.addr, random=False)
            logger.debug("[%s] BD address: %s", self.interface, self._bd_address)
            return True

        # Cannot read BD address, device is non-responsive.
        # logger.error("[%s] cannot read BD address", self.interface)
        logger.debug("cannot read BD address of interface %s", self.interface)
        logger.debug("raising WhadDeviceNotReady exception")
        raise WhadDeviceNotReady(f"cannot read BD address of interface {self.interface}")

    @req_cmd("read_local_name")
    def _read_local_name(self):
        """
        Read local name used by the HCI device.
        """
        logger.debug("[%s] Reading local name ...", self.interface)
        response = self._write_command(HCI_Cmd_Read_Local_Name())
        if response.status == 0x00 and HCI_Cmd_Complete_Read_Local_Name in response:
            self._local_name = response.local_name
            return True

        # Cannot read local name.
        logger.debug("[%s] Failed reading local name !", self.interface)
        logger.debug("[%s] Device not supported.")
        raise WhadDeviceNotReady()

    @req_cmd("read_local_version_information")
    def _read_local_version_information(self):
        """
        Read local version information used by the HCI device.
        """
        logger.debug("[%s] Reading local version info ...", self.interface)
        response = self._write_command(HCI_Cmd_Read_Local_Version_Information())
        if response.status == 0x00 and HCI_Cmd_Complete_Read_Local_Version_Information in response:
            if response.hci_version <= max(list(HCI_VERSIONS.keys())):
                version = [int(v) for v in HCI_VERSIONS[response.hci_version].split(".")]
            else:
                # Unsupported version, default to bluetooth 4.2
                logger.debug(f"Unknown Bluetooth version number {response.hci_version}, falling back to {HCI_VERSIONS[0x06]}.")
                version = [int(v) for v in HCI_VERSIONS[0x08].split(".")]

            version += [response.hci_subversion]
            try:
                manufacturer = BT_MANUFACTURERS[response.company_identifier].encode("utf-8")
            except IndexError:
                logger.debug("[%s] Unsupported manufacturer ID 0x%04x", self.interface,
                             response.company_identifier)
                manufacturer = f"unknown<0x{response.company_identifier:04x}>".encode("utf-8")
            logger.debug("[%s] Version: %s", self.interface, version)
            logger.debug("[%s] Manufacturer: %s", self.interface, manufacturer)
            return version, manufacturer

        # Cannot read local version information.
        logger.debug("[%s] Failed reading local version info !", self.interface)
        logger.debug("[%s] Unsupported HCI interface.")
        raise WhadDeviceNotReady()

    @req_cmd("le_read_supported_states")
    def _read_le_supported_states(self):
        """
        Returns the list of Bluetooth Low Energy states supported by the HCI device.
        """
        logger.debug("[%s] Reading LE supported states ...", self.interface)
        response = self._write_command(HCI_Cmd_LE_Read_Supported_States())
        if response.status == 0x00 and HCI_Cmd_Complete_LE_Read_Supported_States in response:
            states = []
            for bit_position, state in LE_STATES.items():
                if response.supported_states & (1 << bit_position) != 0:
                    states.append(state)
            return states

        # Cannot read supported LE states.
        logger.debug("[%s] Failed reading LE supported states !", self.interface)
        logger.debug("[%s] Unsupported HCI interface.")
        raise WhadDeviceNotReady()

    @req_cmd("le_set_random_address")
    def _set_bd_address(self, bd_address: bytes = b"\x55\x44\x33\x22\x11\x00",
                        bd_address_type: int = AddressType.RANDOM) -> bool:
        """
        Modify the BD address (if supported by the HCI device).
        """
        logger.debug("[%s] Setting HCI adapter random address to %s ...", self.interface, 
                     BDAddress(bd_address))

        # Disabled for now
        if False and bd_address_type == AddressType.PUBLIC:
            _, self._manufacturer = self._read_local_version_information()
            if self._manufacturer in ADDRESS_MODIFICATION_VENDORS:
                logger.info("[i] Address modification supported !")
                if self._manufacturer == b'Qualcomm Technologies International, Ltd. (QTIL)':
                    # Keep in cache existing devices
                    existing_devices = devices = HCIConfig.list()

                    # Write BD address and reset with vendor specific commands
                    self._write_command(HCI_Cmd_CSR_Write_BD_Address(addr=bd_address),
                                        from_queue=False)
                    self._write_command(HCI_Cmd_CSR_Reset(), from_queue=False)

                    # We are forced to close the socket and reopen it here...
                    self.__socket.close()
                    # Add a delay to prevent error
                    sleep(0.5)

                    # The index may have changed, find it automatically and reconfigure self.__index
                    success = False
                    while not success:
                        devices = HCIConfig.list()
                        if self.__index not in devices:
                            for i in existing_devices:
                                if i != self.__index:
                                    devices.remove(i)
                            if len(devices) > 0:
                                self.__index = devices[0]
                                success = True

                    # If all goes right, we should be able to open a new socket
                    self.__socket = get_hci(self.__index)
                    # Initialize a new socket
                    self._initialize()

                else:
                    # For the other manufacturers, we only need to pick the
                    # right command and perform a reset
                    bd_address_mod_map = {
                        b'Texas Instruments Inc.' : HCI_Cmd_TI_Write_BD_Address,
                        b'Broadcom Corporation' : HCI_Cmd_BCM_Write_BD_Address,
                        b'Zeevo, Inc.' : HCI_Cmd_Zeevo_Write_BD_Address,
                        b'Ericsson Technology Licensing' : HCI_Cmd_Ericsson_Write_BD_Address,
                        b'Integrated System Solution Corp.' : HCI_Cmd_Ericsson_Write_BD_Address,
                        b'ST Microelectronics' : HCI_Cmd_ST_Write_BD_Address
                    }
                    command = bd_address_mod_map[self._manufacturer](addr=bd_address)
                    self._write_command(command, from_queue=False)
                    self._reset()

                # Check the modification success and re-generate device ID
                self._bd_address = self._read_bd_address()
                self._dev_id = self._generate_dev_id()
                self._bd_address_type = AddressType.PUBLIC
                return self._bd_address == bd_address

            # Not supported !
            logger.debug("Address modification not supported.")

            # But at least we keep our address type
            self._bd_address_type = bd_address_type
            return False

        if bd_address_type == BDAddress.RANDOM:
            response = self._write_command(HCI_Cmd_LE_Set_Random_Address(address=bd_address))
            if response is not None and response.status == 0x00:
                logger.debug("[%s] Random address successfully set to %s", self.interface, 
                        BDAddress(bd_address))
                # Read BD address
                self._read_bd_address()
                logger.debug("[%s] BD address set to %s (random)", self.interface,
                             BDAddress(bd_address))
                self._bd_address_type = AddressType.RANDOM
            else:
                logger.debug("[%s] Failed setting random address, continue anyway", self.interface)

        # Success
        return True

    @req_cmd("le_set_scan_parameters")
    def _set_scan_parameters(self, active=True):
        """
        Configure Scan parameters for HCI device.
        """
        logger.debug("[%s] Setting LE Scan Parameters (active:%s) ...", self.interface, 
                     active)
        response = self._write_command(HCI_Cmd_LE_Set_Scan_Parameters(type=int(active)))
        return response is not None and response.status == 0x00

    @req_cmd("le_set_scan_enable")
    def _set_scan_mode(self, enable=True):
        """
        Enable or disable scan mode for HCI device.
        """
        logger.debug("[%s] Enabling LE Scan Mode (enable:%s, no duplicates) ...",
                     self.interface, enable)
        response = self._write_command(HCI_Cmd_LE_Set_Scan_Enable(
            enable=int(enable),filter_dups=False
        ))
        return response is not None and response.status == 0x00

    @req_cmd("le_set_host_channel_classification", "le_create_connection")
    def _connect(self, bd_address, bd_address_type=AddressType.PUBLIC, hop_interval=96,
                 channel_map=None):
        """
        Establish a connection using HCI device.
        """
        # Cancel connection if we were trying to connect to a remote peripheral
        if self.__conn_state == HCIConnectionState.ESTABLISHED:
            for handle in self._active_handles:
                self.terminate_connection(handle)
        elif self.__conn_state == HCIConnectionState.INITIATING:
            # Cancel current connection request
            self.cancel_connection()

        logger.debug("bd_address: %s (%d)", bd_address, bd_address_type)
        logger.debug("[hci] _connect() called")
        patype = 0 if bd_address_type == AddressType.PUBLIC else 1
        if channel_map is not None:
            formatted_channel_map = unpack("<Q",channel_map+ b"\x00\x00\x00")[0]
            logger.debug("[%s] Setting Channel Map to  0x%x...",
                self.interface, formatted_channel_map)
            response = self._write_command(HCI_Cmd_LE_Set_Host_Channel_Classification(
                chM=formatted_channel_map
            ))
            if response.status != 0x00:
                logger.debug("[%s] Failed setting Channel Map !", self.interface)
                logger.debug("[%s] Connection aborted.", self.interface)
                return False

        # Connect
        logger.debug("[%s] Creating LE connection to %s ...", self.interface,
                     BDAddress(bd_address, random=bd_address_type==BDAddress.RANDOM))
        response = self._write_command(
            HCI_Cmd_LE_Create_Connection(
                paddr=bd_address,
                patype=patype,
                min_interval=hop_interval,
                max_interval=hop_interval
            )
        )
        if response.status != 0x00:
            # Not connected
            self.__conn_state = HCIConnectionState.DISCONNECTED
            logger.debug("[%s] HCI_LE_Create_Connection command failed with response %d", self.interface,
                         response.status)
        else:
            # Connection is initiating
            self.__conn_state = HCIConnectionState.INITIATING
        return response is not None and response.status == 0x00

    @req_cmd("le_create_connection_cancel")
    def cancel_connection(self) -> bool:
        """When iniating mode, cancel connection creation
        """
        logger.debug("[%s] sending HCI cancel connection command ...", self.interface)
        response = self._write_command(HCI_Cmd_LE_Create_Connection_Cancel())
        if response is not None and response.status == 0x00:
            self.__conn_state = HCIConnectionState.DISCONNECTED
            self._connected = False
            return True
        return False

    @req_cmd("disconnect")
    def _disconnect(self, handle):
        """
        Establish a disconnection using HCI device.
        """
        logger.debug("[%s] sending HCI disconnect command ...")
        response = self._write_command(HCI_Cmd_Disconnect(handle=handle))
        if response is not None and response.status == 0x00:
            self.__conn_state = HCIConnectionState.DISCONNECTED
            self._connected = False
            return True
        return False

    def terminate_connection(self, handle: int):
        """Terminate an active connection or connection attempt.
        """
        if self.__conn_state == HCIConnectionState.INITIATING:
            logger.debug("[%s] HCI interface in connection initiation mode, canceling ...",
                         self.interface)
            if self.cancel_connection():
                logger.debug("[%s] connection initiation successfully canceled.", self.interface)
            else:
                logger.warning("[%s] Cannot cancel pending connection !", self.interface)
        elif self.__conn_state == HCIConnectionState.ESTABLISHED:
            logger.debug("[%s] HCI interface is connected, disconnecting ...")
            if self._disconnect(handle):
                logger.debug("[%s] Successfully disconnected.")
            else:
                logger.warning("[%s] Error while disconnecting !")

    @req_cmd("le_set_advertising_data")
    def _set_advertising_data(self, data, from_queue: bool = True) -> bool:
        """
        Configure advertising data to use by HCI device.
        """
        # pad data if less than 31 bytes
        #if len(data) < 31:
        #    data += b'\x00'*(31 - len(data))

        # Send command and wait for response if required.
        result = True
        if from_queue:
            # Wait for response.
            logger.debug("[%s] Setting HCI LE advertising data ...", self.interface)
            response = self._write_command(HCI_Cmd_LE_Set_Advertising_Data(data=EIR_Hdr(data)))

            # Check response and update result.
            result = response is not None and response.status == 0x0
        else:
            # Otherwise send command without waiting a response.
            logger.debug("[%s] Setting HCI LE advertising data (non-blocking) ...", self.interface)
            self._write_command(HCI_Cmd_LE_Set_Advertising_Data(data=EIR_Hdr(data)), from_queue=False)

        # Return result
        return result

    @req_cmd("le_read_advertising_physical_channel_tx_power")
    def _read_advertising_physical_channel_tx_power(self, from_queue: bool = True) -> bool:
        """Read Advertising Physical Channel Tx Power level
        """
        logger.debug("Read Advertising Physical Channel Tx Power ...")
        response = self._write_command(HCI_Cmd_LE_Read_Advertising_Physical_Channel_Tx_Power(),
                                       from_queue=from_queue)

        if response is not None and response.status == 0x00:
            power_level = response[HCI_Cmd_Complete_LE_Advertising_Tx_Power_Level].tx_power_level
            logger.debug("[%s] Advertising Tx Power level: %d", self.interface, power_level)
            return True
        return False

    @req_cmd("le_set_scan_response_data")
    def _set_scan_response_data(self, data, from_queue=True) -> bool:
        """
        Configure scan response data to use by HCI device.
        """
        result = True
        if from_queue:
            # Wait response and update result accordingly.
            logger.debug("[%s] Setting HCI LE Scan Response Data to %s ...", self.interface,
                         data.hex())
            response = self._write_command(HCI_Cmd_LE_Set_Scan_Response_Data(
                    data=data + (31 - len(data)) * b"\x00", len=len(data)
                )
            )
            result = response is not None and response.status == 0x0
        else:
            # Don't wait for a response.
            logger.debug("[%s] Setting HCI LE Scan Response Data to %s (non-blocking) ...",
                         self.interface, data.hex())
            self._write_command(HCI_Cmd_LE_Set_Scan_Response_Data(
                    data=data + (31 - len(data)) * b"\x00", len=len(data)
                ),
                from_queue=False
            )

        # Return result
        return result

    @req_cmd("le_set_advertising_parameters")
    def set_advertising_parameters(self, interval_min: int = 0x0020, interval_max: int = 0x0020,
                                   adv_type="ADV_IND", oatype: int = 0, datype: int = 0,
                                   daddr:str = "00:00:00:00:00:00", channel_map: int = 0x7,
                                   filter_policy: str = "all:all", from_queue: bool = True) -> bool:
        """Configure the HCI LE advertising parameters.
        """
        # Wait for a response and update result accordingly.
        logger.debug("[%s] Setting HCI LE Advertising Parameters (blocking:%s) ...",
                    self.interface, from_queue)
        response = self._write_command(HCI_Cmd_LE_Set_Advertising_Parameters(
            interval_min = 0x0020,
            interval_max = 0x0020,
            adv_type="ADV_IND",
            oatype=0 if self._bd_address_type == AddressType.PUBLIC else 1,
            datype=0,
            daddr="00:00:00:00:00:00",
            channel_map=0x7,
            filter_policy="all:all"
        ), from_queue=from_queue)

        # Process response if required
        if from_queue:
            return response.status == 0x00
        return True

    @req_cmd("le_set_advertising_enable")
    def _set_advertising_mode(self, enable=True, from_queue=True):
        """
        Enable or disable advertising mode for HCI device.
        """
        # Already enable, nothing to do
        if self._advertising and enable:
            logger.debug("[%s] Advertising is already enabled, nothing to do ...",
                    self.interface)
            return True

        # Not advertising and enabling, we need to configure our advertising
        # parameters.
        if not self._advertising and enable:
            result = self.set_advertising_parameters(
                    interval_min = 0x0020,
                    interval_max = 0x0020,
                    adv_type="ADV_IND",
                    oatype=0 if self._bd_address_type == AddressType.PUBLIC else 1,
                    datype=0,
                    daddr="00:00:00:00:00:00",
                    channel_map=0x7,
                    filter_policy="all:all",
                    from_queue=from_queue
                )

            if result:
                logger.debug("[%s] Sending HCI LE Set Advertise Enable (enable:%s, blocking:%s)",
                             self.interface, enable, from_queue)
                if from_queue:
                    # Wait for a response and update result accordingly.
                    response = self._write_command(HCI_Cmd_LE_Set_Advertise_Enable(enable=int(enable)))
                    result = response.status == 0x00
                else:
                    # Don't wait, simply send command and consider it OK.
                    self._write_command(HCI_Cmd_LE_Set_Advertise_Enable(
                        enable=int(enable)
                    ), from_queue=False)
                    result = True

            # On success, advertising has been enabled.
            self._advertising = result
        else:
            logger.debug("[%s] Sending HCI LE Set Advertise Enable (enable:0, blocking:%s)",
                            self.interface, from_queue)
            # Disabling advertising
            if from_queue:
                response = self._write_command(HCI_Cmd_LE_Set_Advertise_Enable(enable=0))
                result = response.status == 0x00
                if result:
                    self._advertising = False
            else:
                response = self._write_command(HCI_Cmd_LE_Set_Advertise_Enable(enable=0),
                                               from_queue=False)
                self._advertising = False
                result = True

        # Return result
        return result

    @req_cmd("le_long_term_key_request_reply", "le_enable_encryption")
    def _enable_encryption(self, enable=True, handle=None,  key=None, rand=None, ediv=None):
        if self.__converter.pending_key_request:
            logger.debug("[%s] Sending HCI LE Long Term Key request", self.interface)
            response = self._write_command(
                HCI_Cmd_LE_Long_Term_Key_Request_Reply(
                    handle=handle,
                    ltk=key[::-1]
                )
            )
            self.__converter.pending_key_request = False
        else:
            logger.debug("[%s] Sending HCI LE Enable Encryption", self.interface)
            response = self._write_command(
                HCI_Cmd_LE_Enable_Encryption(
                    handle=handle,
                    ltk=key[::-1],
                    rand=rand,
                    ediv=unpack("<H", ediv)[0]
                )
            )
        return response.status == 0x00

    def _update_max_acl_len(self, length: int):
        """Update device HCI MTU
        """
        self.__conn_max_tx_octets = length

    def get_max_acl_len(self) -> int:
        """Retrieve maximum ACL length

        :return: Current HCI maximum TX size
        :rtype: int
        """
        return self.__conn_max_tx_octets


    def _on_whad_ble_encryption(self, message):
        success = self._enable_encryption(
            message.enabled,
            message.conn_handle,
            message.key,
            message.rand,
            message.ediv
        )
        if success:
            self._send_whad_command_result(CommandResult.SUCCESS)
            return
        self._send_whad_command_result(CommandResult.ERROR)

    def on_connection_created(self, handle: int = 0):
        """Callback method to handle a new connection
        """
        # Central mode ?
        if self.__internal_state == HCIInternalState.CENTRAL:
            # HCI interface now connected
            if self.__conn_state == HCIConnectionState.INITIATING:
                # Connection is now established
                self.__conn_state = HCIConnectionState.ESTABLISHED
                self._connected = True
                if handle not in self._active_handles:
                    self._active_handles.append(handle)
                else:
                    logger.warning("[%s] Connection event received with existing handle %d",
                                self.interface, handle)
            else:
                logger.debug("[%s] Unexpected connection event (handle:%d, current state:%d)",
                            self.interface, handle, self.__conn_state)
                logger.warning("[%s] Received an unexpected connection event", self.interface)
        elif self.__internal_state == HCIInternalState.PERIPHERAL:
            # HCI interface now connected
            self.__conn_state = HCIConnectionState.ESTABLISHED
            self._connected = True
            if handle not in self._active_handles:
                self._active_handles.append(handle)
            else:
                logger.warning("[%s] Connection event received with existing handle %d",
                            self.interface, handle)

    def on_connection_terminated(self, handle: int = 0):
        """Callback method to handle connection termination

        :param handle: Connection handle
        :type handle: int
        """
        if self.__conn_state == HCIConnectionState.ESTABLISHED:
            # Connection is now terminated
            self.__conn_state = HCIConnectionState.DISCONNECTED
            self._connected = False

            # Remove handle from active handles
            if handle in self._active_handles:
                self._active_handles.remove(handle)
            else:
                logger.warning("[%s] Disconnect event received for unknown handle %d",
                               self.interface, handle)

    def _on_whad_ble_periph_mode(self, message):
        logger.debug("whad ble periph mode message")
        if Commands.PeripheralMode in self._dev_capabilities[Domain.BtLE][1]:
            success = self._read_advertising_physical_channel_tx_power()
            if len(message.scan_data) > 0:
                success = success and self._set_advertising_data(message.scan_data)
                self._cached_scan_data = message.scan_data
            if len(message.scanrsp_data) > 0:
                success = success and self._set_scan_response_data(message.scanrsp_data)
                self._cached_scan_response_data = message.scanrsp_data
            success = success and self._set_advertising_mode(True)
            if success:
                self.__internal_state = HCIInternalState.PERIPHERAL
                self._send_whad_command_result(CommandResult.SUCCESS)
                return
        else:
            logger.debug("[%s] HCI interface does not allow peripheral mode.")
        self._send_whad_command_result(CommandResult.ERROR)

    def _on_whad_ble_disconnect(self, message):
        success = self._disconnect(message.conn_handle)
        if success:
            self._send_whad_command_result(CommandResult.SUCCESS)
            return
        self._send_whad_command_result(CommandResult.ERROR)

    def _on_whad_ble_connect(self, message):
        logger.debug("[hci] received WHAD connect message")
        if Commands.ConnectTo in self._dev_capabilities[Domain.BtLE][1]:
            bd_address = message.bd_address
            bd_address_type = message.addr_type
            channel_map = message.channel_map if message.channel_map is not None else None
            hop_interval = message.hop_interval if message.hop_interval is not None else 96
            if self._connect(bd_address, bd_address_type, hop_interval=hop_interval,
                             channel_map=channel_map):
                self._send_whad_command_result(CommandResult.SUCCESS)
                return
        self._send_whad_command_result(CommandResult.ERROR)

    def _on_whad_ble_central_mode(self, message):
        if Commands.CentralMode in self._dev_capabilities[Domain.BtLE][1]:
            self.__internal_state = HCIInternalState.CENTRAL
            self._send_whad_command_result(CommandResult.SUCCESS)
            return
        self._send_whad_command_result(CommandResult.ERROR)

    def _on_whad_ble_scan_mode(self, message):
        if Commands.ScanMode in self._dev_capabilities[Domain.BtLE][1]:
            active_scan = message.active
            if self._set_scan_parameters(active_scan):
                self.__internal_state = HCIInternalState.SCANNING
                self._send_whad_command_result(CommandResult.SUCCESS)
                return
        self._send_whad_command_result(CommandResult.ERROR)

    def _on_whad_ble_start(self, _):
        """Enable the current active mode.

        Configure the HCI interface based on selected mode.
        """
        logger.info("whad internal state: %d", self.__internal_state)
        if self.__internal_state == HCIInternalState.SCANNING:
            self._set_scan_mode(True)
            self._send_whad_command_result(CommandResult.SUCCESS)
        elif self.__internal_state == HCIInternalState.PERIPHERAL:
            if not self._advertising:
                if self._set_advertising_mode(True):
                    self._advertising = True
                    self._send_whad_command_result(CommandResult.SUCCESS)
                else:
                    self._send_whad_command_result(CommandResult.ERROR)
            else:
                self._send_whad_command_result(CommandResult.SUCCESS)

        elif self.__internal_state == HCIInternalState.CENTRAL:
            self._send_whad_command_result(CommandResult.SUCCESS)
        else:
            self._send_whad_command_result(CommandResult.ERROR)

    def _on_whad_ble_stop(self, message):

        # Stop any connection attempt, terminate established connections
        if self.__conn_state == HCIConnectionState.INITIATING:
            self.cancel_connection()
        elif self.__conn_state == HCIConnectionState.ESTABLISHED:
            for handle in self._active_handles:
                self.terminate_connection(handle)

        # Process with requested mode
        if self.__internal_state == HCIInternalState.SCANNING:
            self._set_scan_mode(False)
            self.__internal_state = HCIInternalState.NONE
            self._send_whad_command_result(CommandResult.SUCCESS)
        elif self.__internal_state == HCIInternalState.CENTRAL:
            # Update mode and return success
            self.__internal_state = HCIInternalState.NONE
            self._send_whad_command_result(CommandResult.SUCCESS)
        elif self.__internal_state == HCIInternalState.PERIPHERAL:
            # We are not advertising anymore
            self._set_advertising_mode(False)
            self._send_whad_command_result(CommandResult.SUCCESS)
        else:
            self._send_whad_command_result(CommandResult.ERROR)

    def _on_whad_ble_send_pdu(self, message):
        """Send a given PDU into the active connection
        """
        # Make sure we have an active connection
        if self.__conn_state == HCIConnectionState.ESTABLISHED:
            logger.debug("[%s] Received WHAD BLE send_pdu message", self.interface)
            if ((self.__internal_state == HCIInternalState.CENTRAL and message.direction == BleDirection.MASTER_TO_SLAVE) or
            (self.__internal_state == HCIInternalState.PERIPHERAL and message.direction == BleDirection.SLAVE_TO_MASTER)):
                try:
                    hci_packets = self.__converter.process_message(message)

                    if hci_packets is not None:
                        logger.debug("[%s] sending HCI packets ...", self.interface)

                        self.__converter.lock()
                        success = True
                        for hci_packet in hci_packets:
                            success = success and self._write_packet(hci_packet)
                        self.__converter.unlock()
                        logger.debug("[%s] HCI packet sending result: %s", self.interface, success)
                        if success:
                            logger.debug("[%s] send_pdu command succeeded.", self.interface)
                            self._send_whad_command_result(CommandResult.SUCCESS)
                        else:
                            logger.debug("[%s] send_pdu command failed.", self.interface)
                            self._send_whad_command_result(CommandResult.ERROR)

                    pending_messages = self.__converter.get_pending_messages()
                    for pending_message in pending_messages:
                        self._send_whad_message(pending_message)

                except WhadDeviceUnsupportedOperation:
                    logger.debug("Parameter error")
                    self._send_whad_command_result(CommandResult.PARAMETER_ERROR)
            else:
                # Wrong state, cannot send PDU
                logger.debug("wrong state or packet direction.")
                self._send_whad_command_result(CommandResult.ERROR)
        else:
            # Not connected
            logger.debug("[%s] Cannot send PDU: no connection.", self.interface)
            self._send_whad_command_result(CommandResult.ERROR)

    def _on_whad_ble_set_bd_addr(self, message):
        logger.debug("Received WHAD BLE set_bd_addr message")
        if self._set_bd_address(message.bd_address, message.addr_type):
            logger.debug("HCI adapter BD address set to %s", str(message.bd_address))
            self._send_whad_command_result(CommandResult.SUCCESS)
        else:
            logger.debug("HCI adapter does not support BD address spoofing")
            self._send_whad_command_result(CommandResult.ERROR)




"""
Adaptation layer for the Yard Stick One.
"""
import logging
from struct import unpack, pack
from time import time

from usb.core import find, USBError, USBTimeoutError
from usb.util import get_string

from whad.exceptions import WhadDeviceNotFound, WhadDeviceNotReady, WhadDeviceAccessDenied
from whad.device import VirtualDevice
from whad.hub.discovery import Domain, Capability
from whad.hub.generic.cmdresult import CommandResult
from whad.hub.phy import Commands, TxPower, Endianness as PhyEndianness, Modulation as PhyModulation
from whad.phy import Endianness
from whad.helpers import swap_bits

from .constants import YardStickOneId, YardStickOneEndPoints, \
    YardApplications, YardSystemCommands, YardRadioStructure, YardRFStates, \
    YardMemoryRegisters, YardMARCStates, YardCCA, YardFrequencyTransitionPoints, \
    YardNICCommands, YardVCOType, YardRegistersMasks, YardModulations, YardEncodings, \
    POSSIBLE_CHANNEL_BANDWIDTHS, NUM_PREAMBLE_LOOKUP_TABLE, YardUSBProperties, \
    YardInternalStates

logger = logging.getLogger(__name__)

# Helpers functions
def get_yardstickone(index=0,bus=None, address=None):
    '''
    Returns a YardStickOne USB object based on index or bus and address.
    '''
    devices = list(find(idVendor=YardStickOneId.YARD_ID_VENDOR,
                        idProduct=YardStickOneId.YARD_ID_PRODUCT,
                        find_all=True))
    if bus is not None and address is not None:
        for device in devices:
            if device.bus == bus and device.address == address:
                return (devices.index(device), device)
        # No device found with the corresponding bus/address, return None
        return None

    try:
        return (index, devices[index])
    except IndexError:
        return None

class YardStickOne(VirtualDevice):
    """Yard Stick One virtual device.
    """

    INTERFACE_NAME = "yardstickone"

    @classmethod
    def list(cls):
        '''
        Returns a list of available RZUSBStick devices.
        '''
        available_devices = []
        try:
            for yard in find(idVendor=YardStickOneId.YARD_ID_VENDOR,
                             idProduct=YardStickOneId.YARD_ID_PRODUCT,
                             find_all=True):
                available_devices.append(YardStickOne(bus=yard.bus, address=yard.address))
        except ValueError:
            logger.warning("Cannot access YardStickOne, root privileges may be required.")

        return available_devices

    @property
    def identifier(self):
        '''
        Returns the identifier of the current device (e.g., bus + address in
        format "<bus>-<address>").
        '''
        return str(self.__yard.bus)+"-"+str(self.__yard.address)


    def __init__(self, index=0, bus=None, address=None):
        """
        Create device connection
        """
        device = get_yardstickone(index,bus=bus,address=address)
        if device is None:
            raise WhadDeviceNotFound
        self.__opened_stream = False

        self.__supported_frequency_range = [
            (281000000, 361000000),
            (378000000, 481000000),
            (749000000, 962000000),
        ]

        self.radio_structure = None
        self._frequency_offset_accumulator = None
        self.__internal_state = YardInternalStates.YARD_MODE_IDLE
        self.__frequency = None
        self.__endianness = Endianness.BIG
        self._rf_mode = 0

        self.__start_time = time()*1000000

        self.__opened = False
        _, self.__yard = device
        super().__init__()

    @property
    def frequency(self) -> int:
        """Get current frequency.

        :return: Current frequency.
        :rtype: int
        """
        return self.__frequency

    def _enter_configuration_mode(self):
        self._set_idle_mode()
        self._strobe_idle_mode()

    def _restore_previous_mode(self):
        if self.__internal_state == YardInternalStates.YARD_MODE_RX:
            self._set_rx_mode()
            self._strobe_rx_mode()

        elif self.__internal_state == YardInternalStates.YARD_MODE_TX:
            self._set_tx_mode()
            self._strobe_tx_mode()

    def _on_whad_phy_get_supported_freq(self, message):
        # Create a SupportedFreqRanges
        msg = self.hub.phy.create_supported_freq_ranges(
            self.__supported_frequency_range
        )

        # Send message
        self._send_whad_message(msg)

    def _on_whad_phy_send(self, message):
        self._send_packet(message.packet)
        self._send_whad_command_result(CommandResult.SUCCESS)


    def _on_whad_phy_set_freq(self, message):
        found = False
        for supported_range in self.__supported_frequency_range:
            range_start, range_end = supported_range[0], supported_range[1]
            if message.frequency >= range_start and message.frequency <= range_end:
                found = True
                break

        if found:
            self._enter_configuration_mode()
            self._set_frequency(message.frequency)
            self.__frequency = message.frequency
            self._restore_previous_mode()
            self._send_whad_command_result(CommandResult.SUCCESS)
        else:
            self._send_whad_command_result(CommandResult.PARAMETER_ERROR)

    def _on_whad_phy_datarate(self, message):
        self._enter_configuration_mode()
        self._set_data_rate(message.rate)
        self._restore_previous_mode()
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_phy_packet_size(self, message):
        if message.packet_size < 255:
            self._enter_configuration_mode()
            self._set_packet_length(message.packet_size)
            self._restore_previous_mode()
            self._send_whad_command_result(CommandResult.SUCCESS)
        else:
            self._send_whad_command_result(CommandResult.PARAMETER_ERROR)

    def _on_whad_phy_endianness(self, message):
        if message.endianness in (Endianness.BIG, Endianness.LITTLE):
            self.__endianness = message.endianness
            self._send_whad_command_result(CommandResult.SUCCESS)
        else:
            self._send_whad_command_result(CommandResult.PARAMETER_ERROR)

    def _on_whad_phy_tx_power(self, message):
        tx_powers = {
            TxPower.LOW : 0x20,
            TxPower.MEDIUM : 0x80,
            TxPower.HIGH : 0xC0
        }

        if message.power in tx_powers:
            self._enter_configuration_mode()
            self._set_power(
                tx_powers[message.power]
            )
            self._restore_previous_mode()
            self._send_whad_command_result(CommandResult.SUCCESS)
        else:
            self._send_whad_command_result(CommandResult.PARAMETER_ERROR)

    def _on_whad_phy_mod_ask(self, message):
        self._enter_configuration_mode()
        self._set_modulation(YardModulations.MODULATION_ASK)
        self._set_encoding(YardEncodings.NON_RETURN_TO_ZERO)
        self._restore_previous_mode()
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_phy_mod_4fsk(self, message):
        self._enter_configuration_mode()
        self._set_modulation(YardModulations.MODULATION_4FSK)
        self._set_encoding(YardEncodings.NON_RETURN_TO_ZERO)
        self._set_deviation(message.deviation)
        self._restore_previous_mode()
        self._send_whad_command_result(CommandResult.SUCCESS)


    def _on_whad_phy_mod_fsk(self, message):
        self._enter_configuration_mode()
        self._set_modulation(YardModulations.MODULATION_2FSK)
        self._set_encoding(YardEncodings.NON_RETURN_TO_ZERO)
        self._set_deviation(message.deviation)
        self._restore_previous_mode()
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_phy_mod_gfsk(self, message):
        self._enter_configuration_mode()
        self._set_modulation(YardModulations.MODULATION_GFSK)
        self._set_encoding(YardEncodings.NON_RETURN_TO_ZERO)
        self._set_deviation(message.deviation)
        # message.gaussian_filter can't be processed
        self._restore_previous_mode()
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_phy_sync_word(self, message):
        if len(message.sync_word) == 0:
            self._enter_configuration_mode()
            self._set_crc(enable=False)
            self._set_whitening(enable=False)
            self._set_packet_format(0)
            self._set_forward_error_correction(enable=False)
            self._set_clear_channel_assessment(mode=YardCCA.NO_CCA)
            self._set_sync_word(b"")
            self._set_preamble_quality_threshold(0)
            self._restore_previous_mode()
            self._send_whad_command_result(CommandResult.SUCCESS)
        elif len(message.sync_word) <= 2:
            self._enter_configuration_mode()
            self._set_crc(enable=False)
            self._set_whitening(enable=False)
            self._set_packet_format(0)
            self._set_forward_error_correction(enable=False)
            self._set_clear_channel_assessment(mode=YardCCA.NO_CCA)
            if self.__endianness == Endianness.LITTLE:
                sync = bytes([swap_bits(i) for i in message.sync_word])[::-1]
            else:
                sync = message.sync_word
            self._set_sync_word(sync)
            self._set_preamble_quality_threshold(0)
            self._restore_previous_mode()
            self._send_whad_command_result(CommandResult.SUCCESS)
        elif len(message.sync_word) == 4 and message.sync_word[:2] == message.sync_word[2:]:
            self._enter_configuration_mode()
            self._set_crc(enable=False)
            self._set_whitening(enable=False)
            self._set_packet_format(0)
            self._set_forward_error_correction(enable=False)
            self._set_clear_channel_assessment(mode=YardCCA.NO_CCA)
            if self.__endianness == Endianness.LITTLE:
                sync = bytes([swap_bits(i) for i in message.sync_word])[::-1]
            else:
                sync = message.sync_word
            self._set_sync_word(sync)
            self._set_preamble_quality_threshold(0)
            self._restore_previous_mode()
            self._send_whad_command_result(CommandResult.SUCCESS)

        else:
            self._send_whad_command_result(CommandResult.PARAMETER_ERROR)


    def _on_whad_phy_sniff(self, message):
        self.__internal_state = YardInternalStates.YARD_MODE_RX
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_phy_start(self, message):
        if self.__internal_state == YardInternalStates.YARD_MODE_RX:
            self._set_rx_mode()
            self._strobe_rx_mode()
            self.__opened_stream = True
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_phy_stop(self, message):
        #self._set_idle_mode()
        #self._strobe_idle_mode()
        self.__opened_stream = False
        self._send_whad_command_result(CommandResult.SUCCESS)


    def _send_whad_phy_pdu(self, packet, timestamp=None):
        # Create a PacketReceived message
        msg = self.hub.phy.create_packet_received(
            self._get_frequency(),
            packet,
            syncword=self._get_sync_word(),
            endianness=PhyEndianness.LITTLE if self.__endianness == Endianness.LITTLE \
                else PhyEndianness.BIG,
            deviation=int(self._get_deviation()),
            datarate=int(self._get_data_rate())
        )

        # Set packet timestamp if available
        if timestamp is not None:
            msg.timestamp = timestamp

        self._send_whad_message(msg)


    def open(self):
        try:
            self.__yard.set_configuration()
        except USBError as err:
            if err.errno == 13:
                raise WhadDeviceAccessDenied("yardstickone") from err
            raise WhadDeviceNotReady() from err

        self.reset()
        self._dev_id = self._get_serial_number()
        self._fw_author = self._get_manufacturer()
        self._fw_url = self._get_url()
        self._fw_version = self._get_firmware_version()
        self._dev_capabilities = self._get_capabilities()

        self.__opened = True

        self.radio_structure = YardRadioStructure(self._poke, self._peek)

        self._frequency_offset_accumulator = 0
        #print(self.radio_structure)

        self._set_idle_mode()
        self._set_rf_register("MDMCFG4",0x68)
        self._set_rf_register("MDMCFG3",0xb5)
        self._set_rf_register("MDMCFG2",0x80)
        self._set_rf_register("MDMCFG1",0x23)
        self._set_rf_register("MDMCFG0",0x11)
        self._set_modulation(YardModulations.MODULATION_ASK)
        self._set_encoding(YardEncodings.NON_RETURN_TO_ZERO)
        self._set_crc(enable=False)
        self._set_whitening(enable=False)
        self._set_packet_format(0)
        self._set_forward_error_correction(enable=False)
        self._set_clear_channel_assessment(mode=YardCCA.NO_CCA)
        self._set_frequency(433870000)
        self._set_data_rate(10000)
        #self._set_channel_spacing(self.compute_best_channel_bandwidth())
        self._set_channel(0)
        self._set_intermediate_frequency(44444)
        self._set_idle_mode()

        # Ask parent class to run a background I/O thread
        super().open()
        self.__opened_stream = False

    def write(self, data):
        if not self.__opened:
            raise WhadDeviceNotReady()

    def read(self):
        """Read data from Yard Stick One
        """
        if not self.__opened:
            raise WhadDeviceNotReady()

        if self.__opened_stream:
            try:
                self._yard_send_command(YardApplications.NIC, YardNICCommands.SET_RECV_LARGE,
                                        pack("<H", 200))
                data = self._yard_send_command(YardApplications.NIC,YardNICCommands.RECV)
                while self.__opened_stream and self.__internal_state == YardInternalStates.YARD_MODE_RX:

                    data = self._yard_read_response()

                    if data[0] is not None and data[0] != 0xFF and not (len(data[2]) == 1 \
                            and data[2] == b"\xC8"):
                        if self.__endianness == Endianness.LITTLE:
                            formatted_data = bytes([swap_bits(i) for i in data[2]])
                        else:
                            formatted_data = data[2]

                        self._send_whad_phy_pdu(formatted_data, int(
                            time()*1000000 - self.__start_time
                            ))

            except USBTimeoutError:
                pass

    def reset(self):
        self._yard_send_command(
            YardApplications.SYSTEM,
            YardSystemCommands.RESET,
            b"RESET_NOW\x00"
        )

    def close(self):
        if self.__opened:
            self.__opened_stream = False
            self._set_idle_mode()
        super().close()

    # Yard Stick One low level communication primitives

    def _yard_read_response(self, timeout=1000):
        """Read response sent by the yard stick one.
        """
        try:
            response = bytes(self.__yard.read(YardStickOneEndPoints.IN_ENDPOINT,
                                              512, timeout=timeout))
            if len(response) >= 3:
                size = unpack("<H", response[3:5])[0]
                if len(response) >= 5 + size:
                    app = response[1]
                    verb = response[2]
                    data = response[5:5+size]
                    return app, verb, data
        except USBTimeoutError:
            response = (None, None, None)
        return response

    def _yard_send_command(self, app, command, data=b"", timeout=1000, no_response=False):
        message = bytes([app, command]) + pack("<H", len(data)) + data
        if no_response:
            return None
        recv_app, recv_verb, recv_data = None, None, None
        while (recv_app != app and recv_verb != command):
            self.__yard.write(YardStickOneEndPoints.OUT_ENDPOINT, message, timeout=timeout)
            recv_app, recv_verb, recv_data = self._yard_read_response()
        return recv_data

    # Discovery related functions
    def _get_capabilities(self):
        capabilities = {
            Domain.Phy : (
                (Capability.Sniff | Capability.NoRawData),
                [
                    Commands.GetSupportedFrequencies,
                    Commands.SetASKModulation,
                    Commands.SetFSKModulation,
                    Commands.Set4FSKModulation,
                    Commands.GetSupportedFrequencies,
                    Commands.SetFrequency,
                    Commands.SetDataRate,
                    Commands.SetEndianness,
                    Commands.SetTXPower,
                    Commands.SetPacketSize,
                    Commands.SetSyncWord,
                    Commands.Sniff,
                    Commands.Send,
                    Commands.Start,
                    Commands.Stop,
                    Commands.Set4FSKModulation
                ]
            )
        }

        return capabilities

    def _get_manufacturer(self):
        return get_string(self.__yard, self.__yard.iManufacturer).encode('utf-8')

    def _get_serial_number(self):
        """Get serial number.
        """
        return bytes.fromhex(
            self.__yard.serial_number +
            f"{self.__yard.bus:04x}{self.__yard.address:04x}"
        )

    def _get_firmware_version(self):
        response = self._yard_send_command(
            YardApplications.SYSTEM,
            YardSystemCommands.BUILDTYPE
        )
        revision = int(response.split(b" r")[1][:-1])
        return (revision, 0, 0)

    def _get_url(self):
        return "https://github.com/atlas0fd00m/rfcat".encode('utf-8')

    def _send_packet(self, packet, repeat=1, offset=0):
        data = bytes(packet)

        # It is the only solution we found to allow transmitting data without
        # breaking the sniffer mode.
        old_state = self.__internal_state
        self.__internal_state = YardInternalStates.YARD_MODE_TX
        opened_stream = self.__opened_stream
        self.__opened_stream = False

        self._set_idle_mode()
        self._strobe_idle_mode()


        self._set_tx_mode()
        self._strobe_tx_mode()

        self._yard_send_command(YardApplications.NIC, YardNICCommands.XMIT,
                                pack("<HHH", len(data), repeat, offset) + data,
                                timeout=500)

        self.__internal_state = old_state
        self._restore_previous_mode()

        self.__opened_stream = opened_stream


    def _peek(self, address, size):
        return self._yard_send_command(
            YardApplications.SYSTEM,
            YardSystemCommands.PEEK,
            pack("<HH", size, address)
        )

    def _poke(self, address, data=b""):
        return self._yard_send_command(
            YardApplications.SYSTEM,
            YardSystemCommands.POKE,
            pack("<H", address) + data
        )

    def _set_rf_mode(self, mode):
        self._rf_mode = mode
        return self._yard_send_command(
            YardApplications.SYSTEM,
            YardSystemCommands.RFMODE,
            bytes([mode])
        )

    def _set_tx_mode(self):
        self._set_rf_mode(YardRFStates.STX)

    def _set_rx_mode(self):
        self._set_rf_mode(YardRFStates.SRX)

    def _set_idle_mode(self):
        self._set_rf_mode(YardRFStates.SIDLE)

    def _strobe_tx_mode(self):
        self._poke(YardMemoryRegisters.RFST, bytes([YardRFStates.STX]))

    def _strobe_rx_mode(self):
        self._poke(YardMemoryRegisters.RFST, bytes([YardRFStates.SRX]))

    def _strobe_idle_mode(self):
        self._poke(YardMemoryRegisters.RFST, bytes([YardRFStates.SIDLE]))

    def _strobe_cal_mode(self):
        self._poke(YardMemoryRegisters.RFST, bytes([YardRFStates.SCAL]))

    def _strobe_fstxon_mode(self):
        self._poke(YardMemoryRegisters.RFST, bytes([YardRFStates.SFSTXON]))

    def _strobe_return_mode(self):
        self._poke(YardMemoryRegisters.RFST, bytes([self._rf_mode]))

    def _set_rf_register(self, register, value):
        # Update the radio structure
        self.radio_structure.update()
        # Check Main Radio Control State
        old_marc_state = self.radio_structure.get("MARCSTATE")

        # Go back to idle state
        if old_marc_state != YardMARCStates.MARC_STATE_IDLE:
            self._strobe_idle_mode()

        self.radio_structure.set(register, value)

        # Go back to configured mode
        self._strobe_return_mode()


    def _set_clear_channel_assessment(self, mode=YardCCA.CCA_PACKET,
                                      absolute_threshold=0, relative_threshold=1, magnitude=3):

        masks_mcsm1 = YardRegistersMasks.MCSM1
        mcsm1 = self.radio_structure.get("MCSM1") & ~(
            masks_mcsm1.CCA_MODE.mask << masks_mcsm1.CCA_MODE.offset
        )
        mcsm1 |= (mode << masks_mcsm1.CCA_MODE.offset)

        mask_agcctrl2 = YardRegistersMasks.AGCCTRL2
        agcctrl2 = self.radio_structure.get("AGCCTRL2") & ~(
            mask_agcctrl2.MAGN_TARGET.mask << mask_agcctrl2.MAGN_TARGET.offset
        )
        agcctrl2 |= (magnitude << mask_agcctrl2.MAGN_TARGET.offset)

        mask_agcctrl1 = YardRegistersMasks.AGCCTRL1
        agcctrl1 = self.radio_structure.get("AGCCTRL1") & ~(
            (mask_agcctrl1.CARRIER_SENSE_REL_THR.mask << mask_agcctrl1.CARRIER_SENSE_REL_THR.offset) |
            (mask_agcctrl1.CARRIER_SENSE_ABS_THR.mask << mask_agcctrl1.CARRIER_SENSE_ABS_THR.offset)
        )
        agcctrl1 |= (
            ((relative_threshold & mask_agcctrl1.CARRIER_SENSE_REL_THR.mask) << mask_agcctrl1.CARRIER_SENSE_REL_THR.offset) |
            ((absolute_threshold & mask_agcctrl1.CARRIER_SENSE_ABS_THR.mask) << mask_agcctrl1.CARRIER_SENSE_ABS_THR.offset)
        )

        self._set_rf_register("MCSM1", mcsm1)
        self._set_rf_register("AGCCTRL1", agcctrl1)
        self._set_rf_register("AGCCTRL2", agcctrl2)

    def _set_frequency(self, frequency=433920000):
        freq_multiplier = (0x10000 / 1000000.0) / 24
        computed_value = int(frequency * freq_multiplier)

        self._set_rf_register("FREQ2", (computed_value >> 16))
        self._set_rf_register("FREQ1", (computed_value >> 8) & 0xFF)
        self._set_rf_register("FREQ0", (computed_value & 0xFF))

        if (
            (frequency > YardFrequencyTransitionPoints.FREQ_EDGE_900 and
            frequency < YardFrequencyTransitionPoints.FREQ_MID_900) or
            (frequency > YardFrequencyTransitionPoints.FREQ_EDGE_400 and
            frequency < YardFrequencyTransitionPoints.FREQ_MID_400) or
            (frequency < YardFrequencyTransitionPoints.FREQ_MID_300)
        ):
            self._set_rf_register("FSCAL2", YardVCOType.LOW_VCO)
        elif (
            frequency < 1e9 and (
                (frequency > YardFrequencyTransitionPoints.FREQ_MID_900) or
                (frequency > YardFrequencyTransitionPoints.FREQ_MID_400) or
                (frequency > YardFrequencyTransitionPoints.FREQ_MID_300))
            ):
            self._set_rf_register("FSCAL2", YardVCOType.HIGH_VCO)

    def _get_frequency(self):
        freq_multiplier = (0x10000 / 1000000.0) / 24

        self.radio_structure.update()
        num = (
                (self.radio_structure.get("FREQ2") << 16) +
                (self.radio_structure.get("FREQ1") << 8) +
                self.radio_structure.get("FREQ0")
        )
        return int(num / freq_multiplier)

    def _get_frequency_offset_estimate(self):
        self.radio_structure.update()
        frequency_offset_estimate = self.radio_structure.get("FREQEST")
        return int(frequency_offset_estimate)

    def _set_frequency_offset(self, intermediate_frequency_offset):
        self._set_rf_register("FSCTRL0", intermediate_frequency_offset)

    def _get_frequency_offset(self):
        self.radio_structure.update()
        return self.radio_structure.get("FSCTRL0")

    def _adjust_frequency_offset(self):
        self._get_frequency_offset_estimate()
        self._frequency_offset_accumulator += self.radio_structure.get("FREQEST")
        self._frequency_offset_accumulator &= 0xFF
        self._set_frequency_offset(self._frequency_offset_accumulator)

    def _set_modulation(self, modulation, invert=False):
        self.radio_structure.update()
        mask_mdmcfg2 = YardRegistersMasks.MDMCFG2
        mdmcfg2 = self.radio_structure.get("MDMCFG2") & ~(
            mask_mdmcfg2.MOD_FORMAT.mask << mask_mdmcfg2.MOD_FORMAT.offset
        )
        mdmcfg2 |= ((modulation & mask_mdmcfg2.MOD_FORMAT.mask) << mask_mdmcfg2.MOD_FORMAT.offset)

        if modulation == YardModulations.MODULATION_ASK and not invert:
            if self.radio_structure.get("PA_TABLE0") != 0 and self.radio_structure.get("PA_TABLE1") == 0:
                power = self.radio_structure.get("PA_TABLE0")
                self._set_power(power, invert=invert)
        else:
            if self.radio_structure.get("PA_TABLE0") == 0 and self.radio_structure.get("PA_TABLE1") != 0:
                power = self.radio_structure.get("PA_TABLE1")
                self._set_power(power, invert=invert)
        #print(mdmcfg2)
        self._set_rf_register("MDMCFG2", mdmcfg2)

    def _get_modulation(self):
        self.radio_structure.update()
        mask_mdmcfg2 = YardRegistersMasks.MDMCFG2
        format_mod = (
                    self.radio_structure.get("MDMCFG2") &
                    (mask_mdmcfg2.MOD_FORMAT.mask << mask_mdmcfg2.MOD_FORMAT.offset)
        ) >> mask_mdmcfg2.MOD_FORMAT.offset
        return format_mod

    def _set_power(self, power, invert=False):
        modulation = self._get_modulation()
        if modulation == YardModulations.MODULATION_ASK and not invert:
            self.radio_structure.set("PA_TABLE0", 0)
            self.radio_structure.set("PA_TABLE1", power)
            self.radio_structure.set("PA_TABLE2", 0)
            self.radio_structure.set("PA_TABLE3", 0)
            self.radio_structure.set("PA_TABLE4", 0)
            self.radio_structure.set("PA_TABLE5", 0)
            self.radio_structure.set("PA_TABLE6", 0)
            self.radio_structure.set("PA_TABLE7", 0)


        else:
            self.radio_structure.set("PA_TABLE0", power)
            self.radio_structure.set("PA_TABLE1", 0)
            self.radio_structure.set("PA_TABLE2", 0)
            self.radio_structure.set("PA_TABLE3", 0)
            self.radio_structure.set("PA_TABLE4", 0)
            self.radio_structure.set("PA_TABLE5", 0)
            self.radio_structure.set("PA_TABLE6", 0)
            self.radio_structure.set("PA_TABLE7", 0)

        mask_frend0 = YardRegistersMasks.FREND0
        frend0 = self.radio_structure.get("FREND0") & ~(mask_frend0.PA_POWER.mask << mask_frend0.PA_POWER.offset)
        if modulation == YardModulations.MODULATION_ASK:
            frend0 |= (1 << mask_frend0.PA_POWER.offset)
        self._set_rf_register("FREND0", frend0)

    def _get_channel_spacing(self):
        self.radio_structure.update()
        mask_mdmcfg1 = YardRegistersMasks.MDMCFG1
        channel_spacing_mantissa = self.radio_structure.get("MDMCFG0")
        channel_spacing_exponent = (
            (self.radio_structure.get("MDMCFG1") &
            (mask_mdmcfg1.CHANSPC_E.mask << mask_mdmcfg1.CHANSPC_E.offset)
            ) >> mask_mdmcfg1.CHANSPC_E.offset
        )
        channel_spacing = 1000000.0 * (24/(2**18)) * (
            (256 + channel_spacing_mantissa) * (2**channel_spacing_exponent)
        )
        return channel_spacing


    def _set_channel_spacing(self, channel_spacing_khz):
        self.radio_structure.update()
        mask_mdmcfg1 = YardRegistersMasks.MDMCFG1

        for exponent in range(4):
            mantissa = int((channel_spacing_khz * (2**18)) /
                           (((((1000000.0 * 24 * (2**exponent))))-256) +.5))
            if mantissa < 256:
                mdmcfg1 = (
                    self.radio_structure.get("MDMCFG1") &
                    ~(mask_mdmcfg1.CHANSPC_E.mask << mask_mdmcfg1.CHANSPC_E.offset)
                )
                mdmcfg1 |= ((exponent & mask_mdmcfg1.CHANSPC_E.mask) << mask_mdmcfg1.CHANSPC_E.offset)
                self._set_rf_register("MDMCFG1", mdmcfg1)
                self._set_rf_register("MDMCFG0", mantissa)
                return True
        return False

    def _set_packet_length(self, length, variable=False):
        self.radio_structure.update()
        mask_pktctrl0 = YardRegistersMasks.PKTCTRL0
        pktctrl0 = self.radio_structure.get("PKTCTRL0") & ~(
            mask_pktctrl0.LENGTH_CONFIG.mask << mask_pktctrl0.LENGTH_CONFIG.offset
        )
        pktctrl0 |= ((int(variable) & mask_pktctrl0.LENGTH_CONFIG.mask) << mask_pktctrl0.LENGTH_CONFIG.offset)

        self._set_rf_register("PKTLEN", length if length <= 255 else 0)

    def _is_packet_length_variable(self):
        self.radio_structure.update()
        mask_pktctrl0 = YardRegistersMasks.PKTCTRL0
        length_config = (
                        (
                            self.radio_structure.get("PKTCTRL0") &
                            (mask_pktctrl0.LENGTH_CONFIG.mask << mask_pktctrl0.LENGTH_CONFIG.offset)
                        ) >> mask_pktctrl0.LENGTH_CONFIG.offset
        )
        return bool(length_config)

    def _get_packet_length(self):
        self.radio_structure.update()
        return self.radio_structure.get("PKTLEN")

    def _set_crc(self, enable=False):
        self.radio_structure.update()
        mask_pktctrl0 = YardRegistersMasks.PKTCTRL0
        pktctrl0 = self.radio_structure.get("PKTCTRL0") & ~(
            mask_pktctrl0.CRC_EN.mask << mask_pktctrl0.CRC_EN.offset
        )
        pktctrl0 |= int(enable) << mask_pktctrl0.CRC_EN.offset
        self._set_rf_register("PKTCTRL0", pktctrl0)


    def _is_crc_enabled(self):
        self.radio_structure.update()
        mask_pktctrl0 = YardRegistersMasks.PKTCTRL0
        crc_enabled = bool(
            (
                self.radio_structure.get("PKTCTRL0") &
                (mask_pktctrl0.CRC_EN.mask << mask_pktctrl0.CRC_EN.offset)
            ) >> mask_pktctrl0.CRC_EN.offset
        )
        return crc_enabled

    def _set_whitening(self, enable=False):
        self.radio_structure.update()
        mask_pktctrl0 = YardRegistersMasks.PKTCTRL0
        pktctrl0 = self.radio_structure.get("PKTCTRL0") & ~(
            mask_pktctrl0.WHITE_DATA.mask << mask_pktctrl0.WHITE_DATA.offset
        )
        pktctrl0 |= int(enable) << mask_pktctrl0.WHITE_DATA.offset
        self._set_rf_register("PKTCTRL0", pktctrl0)


    def _is_whitening_enabled(self):
        self.radio_structure.update()
        mask_pktctrl0 = YardRegistersMasks.PKTCTRL0
        whitening_enabled = bool(
            (
                self.radio_structure.get("PKTCTRL0") &
                (mask_pktctrl0.WHITE_DATA.mask << mask_pktctrl0.WHITE_DATA.offset)
            ) >> mask_pktctrl0.WHITE_DATA.offset
        )
        return whitening_enabled


    def _set_preamble_quality_threshold(self, threshold=3):
        self.radio_structure.update()
        mask_pktctrl1 = YardRegistersMasks.PKTCTRL1
        pktctrl1 = self.radio_structure.get("PKTCTRL1") & ~(
            mask_pktctrl1.PQT.mask << mask_pktctrl1.PQT.offset
        )
        pktctrl1 |= (threshold & mask_pktctrl1.PQT.mask) << mask_pktctrl1.PQT.offset
        self._set_rf_register("PKTCTRL1", pktctrl1)


    def _get_preamble_quality_threshold(self):
        self.radio_structure.update()
        mask_pktctrl1 = YardRegistersMasks.PKTCTRL1
        pqt = int(
            (
                self.radio_structure.get("PKTCTRL1") &
                (mask_pktctrl1.PQT.mask << mask_pktctrl1.PQT.offset)
            ) >> mask_pktctrl1.PQT.offset
        )
        return pqt


    def _is_append_packet_status_enabled(self):
        self.radio_structure.update()
        mask_pktctrl1 = YardRegistersMasks.PKTCTRL1
        append_packet_status_enabled = bool(
            (
                self.radio_structure.get("PKTCTRL1") &
                (mask_pktctrl1.APPEND_STATUS.mask << mask_pktctrl1.APPEND_STATUS.offset)
            ) >> mask_pktctrl1.APPEND_STATUS.offset
        )
        return append_packet_status_enabled

    def _set_append_packet_status(self, enable=False):
        self.radio_structure.update()
        mask_pktctrl1 = YardRegistersMasks.PKTCTRL1
        pktctrl1 = self.radio_structure.get("PKTCTRL1") & ~(
            mask_pktctrl1.APPEND_STATUS.mask << mask_pktctrl1.APPEND_STATUS.offset
        )
        pktctrl1 |= int(enable) << mask_pktctrl1.APPEND_STATUS.offset
        self._set_rf_register("PKTCTRL1", pktctrl1)


    def _set_encoding(self, encoding=YardEncodings.NON_RETURN_TO_ZERO):
        self.radio_structure.update()
        mask_mdmcfg2 = YardRegistersMasks.MDMCFG2
        mdmcfg2 = self.radio_structure.get("MDMCFG2") & ~(
            mask_mdmcfg2.MANCHESTER_EN.mask << mask_mdmcfg2.MANCHESTER_EN.offset
        )
        mdmcfg2 |= (int(encoding) << mask_mdmcfg2.MANCHESTER_EN.offset)
        self._set_rf_register("MDMCFG2", mdmcfg2)


    def _get_encoding(self):
        self.radio_structure.update()
        mask_mdmcfg2 = YardRegistersMasks.MDMCFG2
        encoding = int(
            (
                self.radio_structure.get("MDMCFG2") &
                (mask_mdmcfg2.MANCHESTER_EN.mask << mask_mdmcfg2.MANCHESTER_EN.offset)
            ) >> mask_mdmcfg2.MANCHESTER_EN.offset
        )
        return YardEncodings.MANCHESTER if encoding else YardEncodings.NON_RETURN_TO_ZERO


    def _set_forward_error_correction(self, enable=False):
        self.radio_structure.update()
        mask_mdmcfg1 = YardRegistersMasks.MDMCFG1
        mdmcfg1 = self.radio_structure.get("MDMCFG1") & ~(
            mask_mdmcfg1.FEC_EN.mask << mask_mdmcfg1.FEC_EN.offset
        )
        mdmcfg1 |= int(enable) << mask_mdmcfg1.FEC_EN.offset
        self._set_rf_register("MDMCFG1", mdmcfg1)


    def _get_forward_error_correction(self):
        self.radio_structure.update()
        mask_mdmcfg1 = YardRegistersMasks.MDMCFG1
        forward_error_correction = bool(
            (
                self.radio_structure.get("MDMCFG1") &
                (mask_mdmcfg1.FEC_EN.mask << mask_mdmcfg1.FEC_EN.offset)
            ) >> mask_mdmcfg1.FEC_EN.offset
        )
        return forward_error_correction


    def _set_intermediate_frequency(self, intermediate_frequency):
        self.radio_structure.update()
        mask_fsctrl1 = YardRegistersMasks.FSCTRL1
        computed_value = int(0.5 + (intermediate_frequency * (2**10)) / (1000000.0 * 24))
        fsctrl1 = self.radio_structure.get("FSCTRL1") & ~(
            mask_fsctrl1.FREQ_IF.mask << mask_fsctrl1.FREQ_IF.offset
        )
        fsctrl1 |= int(computed_value) << mask_fsctrl1.FREQ_IF.offset
        self._set_rf_register("FSCTRL1", fsctrl1)


    def _get_intermediate_frequency(self):
        self.radio_structure.update()
        mask_fsctrl1 = YardRegistersMasks.FSCTRL1
        if_value = int(
            (
                self.radio_structure.get("MDMCFG1") &
                (mask_fsctrl1.FREQ_IF.mask << mask_fsctrl1.FREQ_IF.offset)
            ) >> mask_fsctrl1.FREQ_IF.offset
        )
        intermediate_frequency = (if_value * ((1000000.0 * 24) / (2**10)))
        return intermediate_frequency

    def _set_channel(self, channel):
        self.radio_structure.set("CHANNR", channel)

    def _get_channel(self, channel):
        self.radio_structure.update()
        return self.radio_structure.get("CHANNR")


    def _set_channel_bandwidth(self, bandwidth):
        self.radio_structure.update()

        for exponent in range(4):
            mantissa = int((((24*1000000.0)/(bandwidth*(2**exponent)*8.0)) - 4) + 0.5)
            if mantissa < 4:
                mdmcfg4_mask = YardRegistersMasks.MDMCFG4
                mdmcfg4 = self.radio_structure.get("MDMCFG4") & ~(
                    (mdmcfg4_mask.CHANBW_E.mask << mdmcfg4_mask.CHANBW_E.offset ) |
                    (mdmcfg4_mask.CHANBW_M.mask << mdmcfg4_mask.CHANBW_M.offset )
                )
                mdmcfg4 |= (
                    ((exponent & mdmcfg4_mask.CHANBW_E.mask) << mdmcfg4_mask.CHANBW_E.offset) |
                    ((mantissa & mdmcfg4_mask.CHANBW_M.mask) << mdmcfg4_mask.CHANBW_M.offset)
                )
                self._set_rf_register("MDMCFG4", mdmcfg4)
                bw = 1000.0*24 / (8.0*(4+mantissa) * (2**exponent))
                if bw > 102e3:
                    self._set_rf_register("FREND1", 0xb6)
                else:
                    self._set_rf_register("FREND1", 0x56)

                if bw > 325e3:
                    self._set_rf_register("TEST2", 0x88)
                    self._set_rf_register("TEST1", 0x31)
                else:
                    self._set_rf_register("TEST2", 0x81)
                    self._set_rf_register("TEST1", 0x35)
                return True
        return False

    def _get_channel_bandwidth(self):
        self.radio_structure.update()

        mdmcfg4_mask = YardRegistersMasks.MDMCFG4
        mdmcfg4 = self.radio_structure.get("MDMCFG4")
        exponent = (
            mdmcfg4 & (mdmcfg4_mask.CHANBW_E.mask << mdmcfg4_mask.CHANBW_E.offset )
            ) >> mdmcfg4.CHANBW_E.offset
        mantissa = (
            mdmcfg4 & (mdmcfg4_mask.CHANBW_M.mask << mdmcfg4_mask.CHANBW_M.offset )
            ) >> mdmcfg4.CHANBW_M.offset

        return (1000000.0*24) / (8.0*(4+mantissa) * (2**exponent))

    def _set_data_rate(self, data_rate):
        self.radio_structure.update()
        mdmcfg4 = self.radio_structure.get("MDMCFG4")
        mdmcfg4_mask = YardRegistersMasks.MDMCFG4

        for exponent in range(16):
            mantissa = int(((data_rate * (2**28)) / ((2**exponent) * (24*1000000.0)) - 256) + 0.5)
            if mantissa < 256:
                self._set_rf_register("MDMCFG3", mantissa)
                mdmcfg4 = mdmcfg4 & ~(mdmcfg4_mask.DRATE_E.mask << mdmcfg4_mask.DRATE_E.offset)
                mdmcfg4 |= ((exponent & mdmcfg4_mask.DRATE_E.mask) << mdmcfg4_mask.DRATE_E.offset)
                self._set_rf_register("MDMCFG4", mdmcfg4)
                return True
        return False

    def _get_data_rate(self):
        self.radio_structure.update()
        mdmcfg4_mask = YardRegistersMasks.MDMCFG4
        exponent = int(
            (
                self.radio_structure.get("MDMCFG4") &
                (mdmcfg4_mask.DRATE_E.mask << mdmcfg4_mask.DRATE_E.offset)
            ) >> mdmcfg4_mask.DRATE_E.offset
        )
        mantissa = self.radio_structure.get("MDMCFG3")
        data_rate = 1000000.0 * 24 * (256+mantissa) * (2**exponent) / (2**28)
        return data_rate


    def _set_deviation(self, deviation):
        deviatn_mask = YardRegistersMasks.DEVIATN

        for exponent in range(8):
            mantissa = int(((deviation * (2**17)) / ((2**exponent) * (24*1000000.0)) - 8) + 0.5)
            if mantissa < 8:
                deviatn = (
                    ((mantissa & deviatn_mask.DEVIATION_M.mask) << deviatn_mask.DEVIATION_M.offset) |
                    ((exponent & deviatn_mask.DEVIATION_E.mask) << deviatn_mask.DEVIATION_E.offset)
                )
                self._set_rf_register("DEVIATN", deviatn)
                return True
        return False

    def _get_deviation(self):
        self.radio_structure.update()
        deviatn_mask = YardRegistersMasks.DEVIATN
        deviatn = self.radio_structure.get("DEVIATN")
        exponent = int(
            (
                deviatn &
                (deviatn_mask.DEVIATION_E.mask << deviatn_mask.DEVIATION_E.offset)
            ) >> deviatn_mask.DEVIATION_E.offset
        )
        mantissa = int(
            (
                deviatn &
                (deviatn_mask.DEVIATION_E.mask << deviatn_mask.DEVIATION_E.offset)
            ) >> deviatn_mask.DEVIATION_E.offset
        )

        deviation = 1000000.0 * 24 * (8+mantissa) * (2**exponent) / (2**17)
        return deviation

    def _set_sync_word(self, sync_word, carrier_sense=False, bitflip_tolerance=False):
        sync_mode = 0
        if len(sync_word) == 0:
            sync_mode = int(carrier_sense) << 2 | 0
        elif len(sync_word) == 2:
            sync_mode = (int(carrier_sense) << 2) | (1 << int(bitflip_tolerance))
        elif len(sync_word) == 4:
            if sync_word[:2] == sync_word[2:]:
                sync_mode = (int(carrier_sense) << 2) | 0b11
            else:
                return False
        else:
            return False

        self.radio_structure.update()
        mdmcfg2 = self.radio_structure.get("MDMCFG2")
        mdmcfg2_mask = YardRegistersMasks.MDMCFG2

        mdmcfg2 &= ~(mdmcfg2_mask.SYNC_MODE.mask << mdmcfg2_mask.SYNC_MODE.offset)
        mdmcfg2 |= ((sync_mode & mdmcfg2_mask.SYNC_MODE.mask) <<  mdmcfg2_mask.SYNC_MODE.offset)
        if len(sync_word) > 0:
            self._set_rf_register("SYNC1",sync_word[1])
            self._set_rf_register("SYNC0",sync_word[0])
        self._set_rf_register("MDMCFG2", mdmcfg2)

        return True

    def _get_sync_word(self):
        self.radio_structure.update()
        mdmcfg2 = self.radio_structure.get("MDMCFG2")
        mdmcfg2_mask = YardRegistersMasks.MDMCFG2
        sync_mode = int(
            (mdmcfg2 & (mdmcfg2_mask.SYNC_MODE.mask << mdmcfg2_mask.SYNC_MODE.offset))
            >>  mdmcfg2_mask.SYNC_MODE.offset
        )
        if sync_mode & 0b11 == 0b11:
            multiplier = 2
        elif sync_mode & 0b11 == 0:
            multiplier = 0
        else:
            multiplier = 1
        return multiplier * bytes(
            [self.radio_structure.get("SYNC0"), self.radio_structure.get("SYNC1")]
        )

    def _get_number_of_preamble_bytes(self):
        self.radio_structure.update()
        mdmcfg1 = self.radio_structure.get("MDMCFG1")
        mdmcfg1_mask = YardRegistersMasks.MDMCFG1
        num_preamble_flag = int((
            mdmcfg1 &
            (mdmcfg1.NUM_PREAMBLE.mask << mdmcfg1.NUM_PREAMBLE.offset)
        ) >> mdmcfg1.NUM_PREAMBLE.offset)
        return NUM_PREAMBLE_LOOKUP_TABLE[num_preamble_flag]

    def _set_number_of_preamble_bytes(self, number):
        try:
            num_preamble_flag = NUM_PREAMBLE_LOOKUP_TABLE[number]
        except IndexError:
            return False

        self.radio_structure.update()
        mdmcfg1 = self.radio_structure.get("MDMCFG1")
        mdmcfg1_mask = YardRegistersMasks.MDMCFG1
        mdmcfg1 &= ~(mdmcfg1.NUM_PREAMBLE.mask << mdmcfg1.NUM_PREAMBLE.offset)

        mdmcfg1 |= ((num_preamble_flag & mdmcfg1.NUM_PREAMBLE.mask) << mdmcfg1.NUM_PREAMBLE.offset)
        self._set_rf_register("MDMCFG1", mdmcfg1)
        return True

    def compute_best_deviation(self):
        """Compute best deviation value.
        """
        data_rate = self._get_data_rate()
        if data_rate <= 2400:
            deviation = 5100
        elif data_rate <= 38400:
            deviation = 20000 * ((data_rate - 2400) / 36000)
        else:
            deviation = 129000 * ((data_rate - 38400) / 211600)
        return deviation

    def compute_best_channel_bandwidth(self):
        """Compute best channel bandwidth value.
        """
        frequency = self._get_frequency()
        data_rate = self._get_data_rate()
        #center_frequency = frequency + 14000000
        frequency_uncertainty = 2 * (20e-6 * frequency)
        min_bandwidth = frequency_uncertainty + data_rate
        best_bandwidth = None
        for possible in POSSIBLE_CHANNEL_BANDWIDTHS:
            if min_bandwidth < possible:
                best_bandwidth = possible
                break
        return best_bandwidth

    def _set_power_amplifier_mode(self, mode=0):
        return self._yard_send_command(
            YardApplications.NIC,
            YardNICCommands.SET_AMP_MODE,
            pack("B", mode)
        )
    def _get_power_amplifier_mode(self):
        return self._yard_send_command(
            YardApplications.NIC,
            YardNICCommands.GET_AMP_MODE
        )[0]

    def _set_packet_format(self, format=0):
        self.radio_structure.update()
        mask_pktctrl0 = YardRegistersMasks.PKTCTRL0

        pktctrl0 = self.radio_structure.get("PKTCTRL0") & ~(
            mask_pktctrl0.PKT_FORMAT.mask << mask_pktctrl0.PKT_FORMAT.offset
        )
        pktctrl0 |= ((int(format) & mask_pktctrl0.PKT_FORMAT.mask) << mask_pktctrl0.PKT_FORMAT.offset)

        self._set_rf_register("PKTCTRL0", pktctrl0)

    def set_test_config(self):
        """Put the YardStickOne in test configuration mode.
        """
        self._set_rf_register("IOCFG0",0x06)
        self._set_rf_register("SYNC1",0xaa)
        self._set_rf_register("SYNC0",0xaa)
        self._set_rf_register("PKTLEN",255)
        self._set_rf_register("PKTCTRL1",0x00)
        self._set_rf_register("PKTCTRL0",0x08)
        self._set_rf_register("FSCTRL1",0x0b)
        self._set_rf_register("FSCTRL0",0x00)
        self._set_rf_register("ADDR",0x00)
        self._set_rf_register("CHANNR",0x00)
        self._set_rf_register("MDMCFG4",0x68)
        self._set_rf_register("MDMCFG3",0xb5)
        self._set_rf_register("MDMCFG2",0x80)
        self._set_rf_register("MDMCFG1",0x23)
        self._set_rf_register("MDMCFG0",0x11)
        self._set_rf_register("MCSM2",0x07)
        self._set_rf_register("MCSM1",0x3f)
        self._set_rf_register("MCSM0",0x14)
        self._set_rf_register("DEVIATN",0x45)
        self._set_rf_register("FOCCFG",0x16)
        self._set_rf_register("BSCFG",0x6c)
        self._set_rf_register("AGCCTRL2",0x43)
        self._set_rf_register("AGCCTRL1",0x40)
        self._set_rf_register("AGCCTRL0",0x91)
        self._set_rf_register("FREND1",0x56)
        self._set_rf_register("FREND0",0x10)
        self._set_rf_register("FSCAL3",0xad)
        self._set_rf_register("FSCAL2",0x0A)
        self._set_rf_register("FSCAL1",0x00)
        self._set_rf_register("FSCAL0",0x11)
        self._set_rf_register("TEST2",0x88)
        self._set_rf_register("TEST1",0x31)
        self._set_rf_register("TEST0",0x09)
        self._set_rf_register("PA_TABLE0",0x05)
        self._set_rf_register("PA_TABLE1",0x00)



"""LoRaWAN packets rework
"""

from scapy.packet import Packet, bind_layers
from scapy.fields import StrFixedLenField, BitField, BitEnumField, FCSField, \
    XLEShortField, XLEIntField, LEThreeBytesField, ByteField, LEShortField, \
    StrLenField, XByteField

class EUIField(StrFixedLenField):
    """LoRaWAN Extended Unique Identifier
    """

    def __init__(self, name, default):
        super().__init__(name, default, length=8)

    def i2h(self, pkt, x):
        return ":".join(["{:02x}".format(i) for i in x[::-1]])

    def i2repr(self, pkt, x):
        return self.i2h(pkt, x)
    
    def any2i(self, pkt, x):
        if isinstance(x, str):
            x = bytes.fromhex(x.replace(":",""))[::-1]
        return x

class PHYPayload(Packet):
    """LoRaWAN 1.0 & 1.1 PHYPayload layer
    """
    name = "LoRaWAN PHYPayload"
    fields_desc = [
        BitEnumField("mtype", 0, 3, {
            0: "JoinRequest",
            1: "JoinAccept",
            2: "UnconfirmedDataUp",
            3: "UnconfirmedDataDown",
            4: "ConfirmedDataUp",
            5: "ConfirmedDataDown",
            6: "RejoinRequest",
            7: "Proprietary"
        }),
        BitField("rfu", 0, 3),
        BitField("major", 0, 2),
        FCSField("mic", 0, fmt="<I")
    ]

class JoinRequest(Packet):
    """LoRaWAN Join Request
    """
    name = "LoRaWAN Join Request"
    fields_desc = [
        EUIField("join_eui", "00:00:00:00:00:00:00:00"),
        EUIField("dev_eui", "00:00:00:00:00:00:00:00"),
        XLEShortField("dev_nonce", 0)
    ]

class JoinAccept(Packet):
    """LoRaWAN Join Accept
    """
    name = "LoRaWAN Join Accept"
    fields_desc = [
        LEThreeBytesField("join_nonce", 0),
        LEThreeBytesField("home_netid", 0),
        XLEIntField("dev_addr", 0),
        BitField("rx2_dr", 0, 4),
        BitField("rx1_dr_offset", 0, 3),
        BitEnumField("opt_neg", 0, 1, {0:"v1.0", 1:"v1.1"}),
        ByteField("rx_delay", 0),
    ]


class MACPayloadUplink(Packet):
    """LoRaWAN Uplink MAC Payload
    """
    name = "LoRaWAN Uplink MAC Payload"
    fields_desc = [
        # FHDR fields for uplink payloads
        XLEIntField("dev_addr", 0),
        BitField("adr", 0, 1),
        BitField("adrackreq", 0, 1),
        BitField("ack", 0, 1),
        BitField("classB", 0, 1),
        BitField("fopts_len", 0, 4),
        LEShortField("fcnt", 0),
        StrLenField("fopts", b'', length_from=lambda pkt: pkt.fopts_len),

        # FPort
        XByteField("fport", 0),
    ]

class MACPayloadDownlink(Packet):
    """LoRaWAN Downlink MAC Payload
    """
    name = "LoRaWAN Downlink MAC Payload"
    fields_desc = [
        XLEIntField("dev_addr", 0),
        BitField("adr", 0, 1),
        BitField("adrackreq", 0, 1),
        BitField("ack", 0, 1),
        BitField("classB", 0, 1),
        BitField("fopts_len", 0, 4),
        LEShortField("fcnt", 0),
        StrLenField("fopts", b'', length_from=lambda pkt: pkt.fopts_len),
        
        # FPort
        XByteField("fport", 0),
    ]


class CFList(Packet):
    """LoRaWAN CFList
    """
    name = "LoRaWAN CFList"


bind_layers(PHYPayload, JoinRequest, mtype=0)
bind_layers(PHYPayload, JoinAccept, mtype=1)
bind_layers(JoinAccept, CFList)
bind_layers(PHYPayload, MACPayloadUplink, mtype=2)
bind_layers(PHYPayload, MACPayloadDownlink, mtype=3)
bind_layers(PHYPayload, MACPayloadUplink, mtype=4)
bind_layers(PHYPayload, MACPayloadDownlink, mtype=5)


from .attrlist import gatt_attr_list_iter, GattAttributeDataList, GattHandleUUIDItem, GattHandleItem, \
    GattAttributeValueItem, GattGroupTypeItem
from whad.ble.profile.attribute import UUID


class GattDataHolder(object):
    def __init__(self, **kwargs):
        self.__data = {}
        for arg in kwargs:
            self.__data[arg] = kwargs[arg]

    def __getattr__(self, name):
        if name in self.__data:
            return self.__data[name]
        else:
            raise AttributeError
        

class GattErrorResponse(GattDataHolder):
    def __init__(self, request, handle, reason):
        super().__init__(
            request=request,
            handle=handle,
            reason=reason
        )

class GattFindInfoRequest(GattDataHolder):
    """GATT Find Information Request
    """
    def __init__(self, start: int, end: int):
        super().__init__(
            start=start,
            end=end
        )


class GattFindInfoResponse(GattAttributeDataList):
    """GATT Find Information Response
    """

    FORMAT_HANDLE_UUID_16 = 1
    FORMAT_HANDLE_UUID_128 = 2          

    def __init__(self, format):
        """Store handles list

        :param int format: Handle format
        :param bytes handles_list: Handles data
        """
        self.__format = format
        self.__handles = []
        if self.__format == self.FORMAT_HANDLE_UUID_16:
            super().__init__(4)
        elif self.__format == self.FORMAT_HANDLE_UUID_128:
            super().__init__(18)
        
    @staticmethod
    def from_bytes(format: int, data: bytes):
        if len(data)>1:
            res = GattFindInfoResponse(format)
            item_size = 4 if format == GattFindInfoResponse.FORMAT_HANDLE_UUID_16 else 18
            adl = GattAttributeDataList.from_bytes(data, item_size, GattHandleUUIDItem)
            for item in adl:
                res.append(item)
            return res
        else:
            return None

class GattFindByTypeValueRequest(GattDataHolder):
    def __init__(self, start: int, end: int, attr_type: UUID, attr_value: bytes):
        super().__init__(
            start=start,
            end=end,
            type_uuid=attr_type,
            value=attr_value
        )

class GattFindByTypeValueResponse(GattAttributeDataList):
    def __init__(self):
        super().__init__(4)

    @staticmethod
    def from_bytes(data: bytes):
        res = GattFindByTypeValueResponse()
        for item in gatt_attr_list_iter(4, data):
            res.append(GattHandleItem.from_bytes(item))
        return res

class GattReadByTypeRequest(GattDataHolder):
    def __init__(self, start: int, end: int, attr_type: UUID):
        super().__init__(
            start=start,
            end=end,
            type=attr_type
        )

class GattReadByTypeRequest128(GattDataHolder):
    def __init__(self, start: int, end: int, attr_type1: UUID, attr_type2: UUID):
        super().__init__(
            start=start,
            end=end,
            type1=attr_type1,
            type2=attr_type2
        )


class GattReadByTypeResponse(GattAttributeDataList):
    def __init__(self, item_size):
        super().__init__(item_size)

    @staticmethod
    def from_bytes(item_size: int, data: bytes):
        res = GattReadByTypeResponse(item_size)
        for item in gatt_attr_list_iter(item_size, data):
            res.append(GattAttributeValueItem.from_bytes(item))
        return res

class GattReadRequest(GattDataHolder):
    def __init__(self, handle: int):
        super().__init__(
            handle=handle
        )

class GattReadResponse(GattDataHolder):
    def __init__(self, value: bytes):
        super().__init__(
            value=value
        )

class GattReadBlobRequest(GattDataHolder):
    def __init__(self, handle: int, offset: int):
        super().__init__(
            handle=handle,
            offset=offset
        )

class GattReadBlobResponse(GattDataHolder):
    def __init__(self, value: bytes):
        super().__init__(
            value=value
        )

class GattReadMultipleRequest(GattDataHolder):
    def __init__(self, handles: bytes):
        super().__init__(
            handles=handles
        )

class GattReadMultipleResponse(GattDataHolder):
    def __init__(self, values: bytes):
        super().__init__(
            values=values
        )

class GattReadByGroupTypeRequest(GattDataHolder):
    def __init__(self, start: int, end: int, group_type: UUID):
        super().__init__(
            start=start,
            end=end,
            type=group_type
        )

class GattReadByGroupTypeResponse(GattAttributeDataList):
    """GATT Read By Group Type Response
    """

    def __init__(self, item_size: int):
        super().__init__(item_size)

    @staticmethod
    def from_bytes(item_size, data):
        res = GattReadByGroupTypeResponse(item_size)
        for item in gatt_attr_list_iter(item_size, data):
            res.append(GattGroupTypeItem.from_bytes(item))
        return res

class GattWriteRequest(GattDataHolder):
    def __init__(self, handle: int, value: bytes):
        super().__init__(
            handle=handle,
            value=value
        )

class GattWriteResponse(GattDataHolder):
    def __init__(self):
        super().__init__()

class GattWriteCommand(GattDataHolder):
    def __init__(self, handle: int, value: bytes):
        super().__init__(
            handle=handle,
            value=value
        )

class GattPrepareWriteRequest(GattDataHolder):
    def __init__(self, handle: int, offset: int, value: bytes):
        super().__init__(
            handle=handle,
            offset=offset,
            value=value
        )

class GattPrepareWriteResponse(GattDataHolder):
    def __init__(self, handle: int, offset: int, value: bytes):
        super().__init__(
            handle=handle,
            offset=offset,
            value=value
        )

class GattExecuteWriteRequest(GattDataHolder):
    def __init__(self, flags: int):
        super().__init__(flags=flags)

class GattExecuteWriteResponse(GattDataHolder):
    def __init__(self):
        super().__init__()

class GattHandleValueNotification(GattDataHolder):
    def __init__(self, handle: int, value: bytes):
        super().__init__(
            handle=handle,
            value=value
        )

class GattHandleValueIndication(GattDataHolder):
    def __init__(self, handle: int, value: bytes):
        super().__init__(
            handle=handle,
            value=value
        )

class GattHandleValueConfirmation(GattDataHolder):
    def __init__(self):
        super().__init__()

class GattExchangeMtuResponse(GattDataHolder):
    def __init__(self, mtu: int):
        super().__init__(
            mtu=mtu
        )

class GattExchangeMtuRequest(GattDataHolder):
    def __init__(self, mtu: int):
        super().__init__(
            mtu=mtu
        )



from whad.zigbee.stack.apl.zcl.exceptions import ZCLCommandNotFound

class ZCLCommand:
    """
    This class represents a Zigbee Cluster Library command.
    """
    def __init__(
                    self,
                    name,
                    generate_callback=None,
                    receive_callback=None
    ):
        self.name = name
        self.generate_callback = generate_callback
        self.receive_callback = receive_callback

class ZCLCommands:
    """
    This class represents a database of Zigbee Cluster Library commands.
    """
    def __init__(self):
        self.commands = {}

    def add_command(self, id, name, generate_callback=None, receive_callback=None):
        """
        Add a command in the command database.
        """
        self.commands[id] = ZCLCommand(
            name,
            generate_callback,
            receive_callback
        )

    def get_command_by_id(self, id):
        """
        Get a command from the database according to its identifier.
        """
        if id in self.commands:
            return self.commands[id]
        raise ZCLCommandNotFound()

    def get_command_by_callback(self, callback):
        """
        Get a command from the database according to one of its callbacks.
        """
        for command_id, command in self.commands.items():
            if (
                command.generate_callback == callback or
                command.receive_callback == callback
            ):
                return (command_id, command)
                
        raise ZCLCommandNotFound()



from typing import Union
from dataclasses import dataclass, fields


def channel_to_frequency(channel):
    '''
    Converts 802.15.4 channel to frequency (in Hz).
    '''
    return 1000000 * (2405 + 5 * (channel - 11))

@dataclass(repr=False)
class Metadata:
    raw : bool = None
    decrypted : bool = None
    timestamp : Union[int, float] = None
    channel : int = None
    rssi : int = None

    def convert_to_header(self):
        pass

    def __repr__(self):
        metadatas = []
        for field in fields(self.__class__):
            if hasattr(self, field.name) and getattr(self,field.name) is not None:
                metadatas.append("{}={}".format(field.name, getattr(self,field.name)))

        if len(metadatas) == 0:
            return ""
        else:
            return "[ " + ", ".join(metadatas) + " ]"



"""
Write procedure
"""
from typing import List

from scapy.packet import Packet
from scapy.layers.bluetooth import ATT_Error_Response, ATT_Handle_Value_Indication, ATT_Handle_Value_Notification, ATT_Write_Request, ATT_Write_Response, ATT_Hdr

from whad.ble.stack.att.constants import BleAttErrorCode, BleAttOpcode

from .attribute import Characteristic, ClientCharacteristicConfigurationDescriptor, find_attr_by_handle, find_charac_by_desc_handle
from .procedure import BleClientProcedure, BleServerProcedure, UnexpectedProcError


class ServerWriteProcedure(BleServerProcedure):
    """ATT server Write procedure."""

    OPCODE = BleAttOpcode.WRITE_REQUEST

    def __init__(self, attributes: list, mtu: int):
        """Initialize our ServerWrite procedure."""
        super().__init__(attributes, mtu)

    @classmethod
    def trigger(cls, request) -> bool:
        """Determine if the procedure is triggered."""
        return ATT_Write_Request in request

    def process_request(self, request: Packet) -> List[Packet]:
        """React only on WriteRequest."""
        if ATT_Write_Request not in request and self.get_state() == self.states.INITIAL:
            self.set_state(self.states.ERROR)
            raise UnexpectedProcError()
        # If we receive a confirmation then indication has correctly been sent.
        elif ATT_Hdr in request and request[ATT_Hdr].opcode == 0x1e and self.get_state() == self.states.INDICATION_SENT:
            self.set_state(self.states.DONE)
            return []

        request = request[ATT_Write_Request]

        # Search for the requested attribute
        try:
            attr = find_attr_by_handle(self.attributes, request.gatt_handle)

            # Find parent characteristic
            charac = find_charac_by_desc_handle(self.attributes, request.gatt_handle)
            if charac is None:
                self.set_state(self.states.DONE)
                return [ ATT_Write_Response() ]

            # Make sure the characteristic has the notification or indication bit set
            if request.data == b"\x01\x00" and not charac.has_property(Characteristic.PROP_NOTIFY):
                return self.att_error_response(request.gatt_handle, BleAttErrorCode.VALUE_NOT_ALLOWED)
            if request.data == b"\x02\x00" and not charac.has_property(Characteristic.PROP_INDICATE):
                return self.att_error_response(request.gatt_handle, BleAttErrorCode.VALUE_NOT_ALLOWED)

            # Check we are attempting to modify a characteristic value or a CCCD
            if attr.writeable():
                attr.value = request.data
                # If we are writing into a CCCD, we must send a notification/indication immediately after, and
                # wait for a confirmation if required.
                if isinstance(attr, ClientCharacteristicConfigurationDescriptor):

                    # Check if we enabled notification or indication
                    if attr.value[0] == 0x01:
                        # We must find the parent characteristic and send a notification
                        notification = ATT_Handle_Value_Notification(gatt_handle=charac.value_handle, value=b"Notified")
                        self.set_state(self.states.DONE)
                        return [ ATT_Write_Response(), notification ]
                    elif attr.value[0] == 0x02:
                        # Find the parent characteristic and send an indication
                        indication = ATT_Handle_Value_Indication(gatt_handle=charac.value_handle, value=b"Indicated")
                        self.set_state(self.states.INDICATION_SENT)
                        return [ ATT_Write_Response(), indication ]
                    else:
                        self.set_state(self.states.DONE)
                        return [ ATT_Write_Response() ]
                else:
                    self.set_state(self.states.DONE)
                    return [ ATT_Write_Response() ]
            else:
                self.set_state(self.states.DONE)
                return self.att_error_response(request.gatt_handle, BleAttErrorCode.WRITE_NOT_PERMITTED)
        except IndexError:
            self.set_state(self.states.DONE)
            return self.att_error_response(request.gatt_handle, BleAttErrorCode.ATTRIBUTE_NOT_FOUND)

class ClientWriteProcedure(BleClientProcedure):
    """GATT Client Write procedure emulation."""

    OPCODE = BleAttOpcode.WRITE_RESPONSE

    def __init__(self, handle: int, value: bytes):
        """Initialize a GATT client Write procedure."""
        self.__handle = handle
        self.__value = value

        # Call parent init method with initial packets to
        # be sent.
        super().__init__([
            ATT_Write_Request(
                gatt_handle=self.__handle,
                data=self.__value,
            )
        ])

    def process_request(self, request: Packet) -> List[Packet]:
        """Process incoming PDU."""
        if ATT_Error_Response in request:
            self.set_result(request[ATT_Error_Response])
            self.set_state(self.states.ERROR)
        elif ATT_Hdr in request and request[ATT_Hdr].opcode == BleAttOpcode.WRITE_RESPONSE:
            self.set_result(True)
            self.set_state(self.states.DONE)
        else:
            print("Error !")
            self.set_result(None)
            self.set_state(self.states.ERROR)
        return []




"""Generic Access Profile service

TODO: Add appearance interpretation.
"""

from whad.ble.profile.attribute import UUID
from whad.ble.profile.service import StandardService
from whad.ble.profile.characteristic import Characteristic

class GapService(StandardService):

    _uuid = UUID(0x1800)

    device_name = Characteristic(
        UUID(0x2a00),
        properties=Characteristic.READ,
        required=True,
        value=b''
    )

    appearance = Characteristic(
       UUID(0x2a01),
        properties=Characteristic.READ,
        required=True,
        value=b'\x00\x00' # Unknown by default
    )

    preferred_conn_params = Characteristic(
        UUID(0x2a04),
        properties=Characteristic.READ,
        value=b'\x06\x00\xa0\x00\x06\x00\xe8\x03'
    )

    @property
    def devicename(self) -> str:
        return self.device_name.value.decode('utf-8')



from scapy.packet import Packet, bind_layers
from scapy.fields import StrField, StrFixedLenField, XByteField, \
    FieldLenField, BitField, BitEnumField, XShortField, StrLenField
from scapy.config import conf
from struct import pack, unpack

from whad.helpers import bits_to_bytes, bytes_to_bits

USER_DLT = 148

def crc_update(crc, value, bits):
    """Update CRC with the provided bits.

    Implementation taken from Bastille's nRF Research firmware
    (https://github.com/BastilleResearch/nrf-research-firmware/)
    """
    crc = crc ^ (value << 8)
    while bits>0:
        bits -= 1
        if (crc & 0x8000) == 0x8000:
            crc = (crc << 1) ^ 0x1021
        else:
            crc = crc << 1
    crc = crc & 0xFFFF
    return crc


def compute_crc(packet) -> bytes:
    """Compute ESB packet CRC bytes (16-bit)

    :param packet: ESB packet with an extra null byte
    :type packet: :class:`whad.scapy.layers.esb.ESB_Hdr`
    :return: CRC encoded as a byte array
    :rtype: bytes
    """
    # CRC initial value is 0xFFFF
    crc = 0xFFFF

    # Loop on all packet bytes except the last one
    for x in packet[:-1]:
        crc = crc_update(crc, x, 8)

    # Include the last bit stored in the extra byte
    crc = crc_update(crc, packet[-1], 1)

    # Return the encoded CRC (big-endian)
    return pack('>H', crc)


# Field representing a ShockBurst address
class SBAddressField(StrLenField):
    def __init__(self, name, default, length_from):
        StrLenField.__init__(self, name, default,length_from=length_from)

    def i2h(self,pkt,x):
        return ":".join(["{:02x}".format(i) for i in x])

    def i2repr(self,pkt,x):
        return self.i2h(pkt, x)

    def any2i(self,pkt,x):
        if isinstance(x,str):
            x = bytes.fromhex(x.replace(":",""))
        return x

class ESB_Hdr(Packet):
    ESB_PREAMBLE_SIZE = 8
    ESB_PCF_SIZE = 9
    ESB_CRC_SIZE = 16
    ESB_PAYLEN_SIZE = 6

    name = "Enhanced ShockBurst packet"
    fields_desc = [
            XByteField("preamble",0xAA),
            FieldLenField("address_length", None, length_of="address"),
            SBAddressField("address",b"\0\0\0\0\0",length_from = lambda pkt:pkt.address_length),
            BitField("payload_length", None, 6),
            BitField("pid",None,2),
            BitField("no_ack", 0,1),
            BitField("padding",0,6),
            BitEnumField("valid_crc",0,1,{0:"no",1:"yes"}),
            XShortField("crc",None)
    ]


    def post_build(self, p, pay):
        """Effectively build the ESB packet based on its properties.
        """
        # Extract preamble
        preamble = p[0]

        # extract address and address length
        address_len = unpack('>H', p[1:3])[0]
        address = p[3:3+address_len]

        # Compute PCF field (9 bits)
        pcf = (p[3+address_len]<<1) | (p[3+address_len+1]&0x80)>>7

        # Update payload length if required
        if self.payload_length is None:
            payload_length = len(pay)
            pcf = (pcf & 0x7) | (payload_length << 3)

        # Build payload bytes
        carry = pcf&1
        out = []
        for x in pay:
            out.append((x>>1) | (carry << 7))
            carry = x&1
        out.append(carry<<7)

        # Compute CRC
        frame = address + bytes([pcf>>1]) + bytes(out)
        if self.crc is None:
            crc = compute_crc(frame)#pack('>H', compute_crc_value(frame))
        else:
            crc = p[-2:]

        # Append the packet CRC
        out[-1] |= crc[0]>>1
        out.append((crc[1]>>1) | (crc[0]&1)<<7)
        out.append((crc[1]&1)<<7)


        # Build the final frame
        return bytes([preamble]) + address + bytes([pcf>>1]) + bytes(out)

    def post_dissect(self, s):
        """Override layer post_dissect() function to reset raw packet cache.
        """
        self.raw_packet_cache = None  # Reset packet to allow post_build
        return s

    def pre_dissect(self,s):
        """Pre-dissect an ESB frame to guess the address and payload sizes.
        """
        if s[0] not in [0xAA, 0x55]: # Dirty patch if no preamble is included
            s = b"\xAA"+s
        bitstring = bytes_to_bits(s)
        crc = None
        crc_found = False
        i = ESB_Hdr.ESB_PREAMBLE_SIZE+1
        # We try to guess the packet size by looking for a valid CRC
        while i < len(bitstring) - 16:
            if bytes_to_bits(compute_crc(bits_to_bytes(bitstring[ESB_Hdr.ESB_PREAMBLE_SIZE:i]))) == bitstring[i:i+ESB_Hdr.ESB_CRC_SIZE]:
                crc_found = True
                break
            i += 1

        # We try to guess the address size by checking if :
        # ESB_PREAMBLE_SIZE + 8*addr_size + ESB_PCF_SIZE + payload_size = 8*packet_size - ESB_CRC_SIZE
        addr_len_found = False
        for addr_length in range(3,6):
            payLen = bits_to_bytes("00"+bitstring[ESB_Hdr.ESB_PREAMBLE_SIZE+addr_length*8:ESB_Hdr.ESB_PREAMBLE_SIZE+addr_length*8+ESB_Hdr.ESB_PAYLEN_SIZE])[0]
            if ESB_Hdr.ESB_PREAMBLE_SIZE+addr_length*8+ESB_Hdr.ESB_PCF_SIZE+payLen*8 == i:
                addr_len_found = True
                break


        preamble = bitstring[:ESB_Hdr.ESB_PREAMBLE_SIZE]
        if crc_found and addr_len_found:
            # No problem, we know that the packet is valid
            address = bitstring[ESB_Hdr.ESB_PREAMBLE_SIZE:ESB_Hdr.ESB_PREAMBLE_SIZE+addr_length*8]
            validCrc = "1" if crc_found else "0"
        else:
            # Our assumption is : addrLen = 5, invalid CRC
            addr_length = 5
            address = bitstring[ESB_Hdr.ESB_PREAMBLE_SIZE:ESB_Hdr.ESB_PREAMBLE_SIZE+addr_length*8]
            validCrc = "0"

        pcf = bitstring[ESB_Hdr.ESB_PREAMBLE_SIZE+addr_length*8:ESB_Hdr.ESB_PREAMBLE_SIZE+addr_length*8+ESB_Hdr.ESB_PCF_SIZE]
        payload_length = bits_to_bytes("00"+pcf[:6])[0]
        payload = bitstring[ESB_Hdr.ESB_PREAMBLE_SIZE+addr_length*8+ESB_Hdr.ESB_PCF_SIZE:ESB_Hdr.ESB_PREAMBLE_SIZE+addr_length*8+ESB_Hdr.ESB_PCF_SIZE+payload_length*8]
        crc = bitstring[ESB_Hdr.ESB_PREAMBLE_SIZE+addr_length*8+ESB_Hdr.ESB_PCF_SIZE+payload_length*8:ESB_Hdr.ESB_PREAMBLE_SIZE+addr_length*8+ESB_Hdr.ESB_PCF_SIZE+payload_length*8+ESB_Hdr.ESB_CRC_SIZE]

        padding = "0"*6

        #print("pl",bits_to_bytes(preamble + bytes_to_bits(bytes([0,addr_length])) + address + pcf + padding + validCrc + crc + payload).hex())
        return bits_to_bytes(preamble + bytes_to_bits(bytes([0,addr_length])) + address + pcf + padding + validCrc + crc + payload)


class ESB_Payload_Hdr(Packet):
    """ESB payload header
    """
    name = "ESB Payload"
    fields_desc = []

class ESB_Ping_Request(Packet):
    """ESB Ping request.

    Ping request is identified by a payload of [0x0F, 0x0F, 0x0F, 0x0F].
    """
    name = "ESB Ping Request"
    fields_desc = [StrFixedLenField('ping_payload', '\x0f\x0f\x0f\x0f', length=4)]

class ESB_Ack_Response(Packet):
    """ESB Ack response layer.
    """
    name = "ESB Ack Response"
    fields_desc = [StrField('ack_payload', '')]


class ESB_Pseudo_Packet(Packet):
    """ESB pseudo-packet layer.
    """
    name = "ESB Pseudo packet"
    fields_desc = []

def guess_payload_class_esb(self, payload):
    """Guess payload content based on payload size and content.
    """
    if b"\x0f\x0f\x0f\x0f" == payload[:4]:
        return ESB_Ping_Request
    elif len(payload) == 0 or self.underlayer is not None and self.underlayer.no_ack == 1:
        return ESB_Ack_Response
    else:
        return Packet.guess_payload_class(self, payload)

ESB_Payload_Hdr.guess_payload_class = guess_payload_class_esb

bind_layers(ESB_Hdr,ESB_Payload_Hdr)
conf.l2types.register(USER_DLT, ESB_Hdr)
conf.l2types.register(USER_DLT, ESB_Pseudo_Packet)



"""
RZUSBStick adaptation layer.
"""
import logging
from struct import unpack, pack

from usb.core import find, USBError, USBTimeoutError

from whad.exceptions import WhadDeviceNotFound, WhadDeviceNotReady, WhadDeviceAccessDenied
from whad.device import VirtualDevice
from whad.hub.discovery import Domain, Capability

from whad.hub.generic.cmdresult import CommandResult
from whad.hub.dot15d4 import Commands

from .constants import RZUSBStickInternalStates, \
    RZUSBStickId, RZUSBStickModes, RZUSBStickEndPoints, RZUSBStickCommands, \
    RZUSBStickResponses


logger = logging.getLogger(__name__)

# Helpers functions
def get_rzusbstick(index=0,bus=None, address=None):
    """
    Returns a RZUSBStick USB object based on index or bus and address.
    """
    devices = list(find(idVendor=RZUSBStickId.RZUSBSTICK_ID_VENDOR,
                        idProduct=RZUSBStickId.RZUSBSTICK_ID_PRODUCT,
                        find_all=True))
    if bus is not None and address is not None:
        for device in devices:
            if device.bus == bus and device.address == address:
                return (devices.index(device), device)
        # No device found with the corresponding bus/address, return None
        return None

    try:
        return (index, devices[index])
    except IndexError:
        return None

class RzUsbStick(VirtualDevice):
    """ATMEL RZAVRUSB stick virtual device.
    """

    INTERFACE_NAME = "rzusbstick"

    @classmethod
    def list(cls):
        """
        Returns a list of available RZUSBStick devices.
        """
        available_devices = []
        try:
            for rzusbstick in find(idVendor=RZUSBStickId.RZUSBSTICK_ID_VENDOR,
                                   idProduct=RZUSBStickId.RZUSBSTICK_ID_PRODUCT,
                                   find_all=True):
                available_devices.append(RzUsbStick(bus=rzusbstick.bus,
                                                          address=rzusbstick.address))
        except ValueError:
            logger.warning("Cannot access RZUSBStick, root privileges may be required.")

        return available_devices

    @property
    def identifier(self):
        """
        Returns the identifier of the current device (e.g., bus + address in
        format "<bus>-<address>").
        """
        return str(self.__rzusbstick.bus)+"-"+str(self.__rzusbstick.address)


    def __init__(self, index=0, bus=None, address=None):
        """
        Create device connection
        """
        device = get_rzusbstick(index,bus=bus,address=address)
        if device is None:
            raise WhadDeviceNotFound

        self.__input_header = None
        self.__input_buffer = None
        self.__input_buffer_length = 0
        self.__opened = False
        self.__opened_stream = False
        self.__channel = 11
        self.__future_channel = 11
        self.__internal_state = RZUSBStickInternalStates.NONE
        _, self.__rzusbstick = device
        super().__init__()

    @property
    def internal_state(self):
        """Get the device internal state.
        """
        return self.__internal_state

    def open(self):
        try:
            self.__rzusbstick.set_configuration()
        except USBError as err:
            if err.errno == 13:
                raise WhadDeviceAccessDenied("rzusbstick") from err
            raise WhadDeviceNotReady() from err

        self._dev_id = self._get_serial_number()
        self._fw_author = self._get_manufacturer()
        self._fw_url = self._get_url()
        self._fw_version = self._get_firmware_version()
        self._dev_capabilities = self._get_capabilities()

        self.__opened = True
        # Ask parent class to run a background I/O thread
        super().open()

    def write(self, data):
        if not self.__opened:
            raise WhadDeviceNotReady()

    def read(self):
        """Read data from device.
        """
        if not self.__opened:
            raise WhadDeviceNotReady()
        if self.__opened_stream:
            try:
                data = self.__rzusbstick_read_packet()
            except USBTimeoutError:
                data = b""
            if data is not None and len(data) >= 1:
                if data[0] == RZUSBStickResponses.RZ_AIRCAPTURE_DATA and len(data) >= 10:
                    self.__input_header = data[:9]
                    self.__input_buffer = data[9:]
                    self.__input_buffer_length = data[1] - 9
                elif len(self.__input_header) > 0:
                    self.__input_buffer += data
                else:
                    self.__input_buffer = b""
                    self.__input_buffer_length = 0
                    self.__input_header = b""

                if len(self.__input_header) == 9 and self.__input_buffer_length == len(self.__input_buffer):
                    rssi = 3 * self.__input_header[6] - 91
                    valid_fcs = self.__input_header[7] == 0x01
                    packet = self.__input_buffer[:-1]
                    #link_quality_indicator = self.__input_buffer[-1]
                    self._send_whad_zigbee_raw_pdu(packet, rssi=rssi, is_fcs_valid=valid_fcs)

    def reset(self):
        self.__rzusbstick.reset()

    # Virtual device whad message builder
    def _send_whad_zigbee_raw_pdu(self, packet, rssi=None, is_fcs_valid=None, timestamp=None):
        pdu = packet[:-2]
        fcs = unpack("H",packet[-2:])[0]

        # Create a RawPduReceived message
        msg = self.hub.dot15d4.create_raw_pdu_received(
            self.__channel,
            pdu,
            fcs,
            fcs_validity=is_fcs_valid
        )

        # Set timestamp and RSSI if provided
        if rssi is not None:
            msg.rssi = rssi
        if timestamp is not None:
            msg.timestamp = timestamp

        # Send message

        self._send_whad_message(msg)


    # Virtual device whad message callbacks
    def _on_whad_dot15d4_stop(self, message):
        if self._stop():
            self._send_whad_command_result(CommandResult.SUCCESS)
        else:
            self._send_whad_command_result(CommandResult.ERROR)

    def _on_whad_dot15d4_send_raw(self, message):
        channel = message.channel

        if self._set_channel(channel):
            packet = message.pdu + pack("H",message.fcs)
            success = self._send_packet(packet)
        else:
            success = False
        self._send_whad_command_result(CommandResult.SUCCESS if success else CommandResult.ERROR)

    def _on_whad_dot15d4_sniff(self, message):
        channel = message.channel
        self.__future_channel = channel
        self.__internal_state = RZUSBStickInternalStates.SNIFFING
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_dot15d4_start(self, message):
        self.__input_buffer = b""
        self.__input_buffer_length = 0
        self.__input_header = b""
        if self._start():
            if self.__future_channel != self.__channel:
                self._set_channel(self.__future_channel)
            self._send_whad_command_result(CommandResult.SUCCESS)
        else:
            self._send_whad_command_result(CommandResult.ERROR)

    # RZUSBStick low level communication primitives

    def __rzusbstick_read_packet(self, timeout=200):
        return bytes(self.__rzusbstick.read(RZUSBStickEndPoints.RZ_PACKET_ENDPOINT,
                                            self.__rzusbstick.bMaxPacketSize0, timeout=timeout))

    def _rzusbstick_read_response(self, timeout=200):
        return bytes(self.__rzusbstick.read(RZUSBStickEndPoints.RZ_RESPONSE_ENDPOINT,
                                            self.__rzusbstick.bMaxPacketSize0, timeout=timeout))

    def _rzusbstick_send_command(self, command, data=b"", timeout=200):
        data = [command] + list(data)
        self.__rzusbstick.write(RZUSBStickEndPoints.RZ_COMMAND_ENDPOINT, data, timeout=timeout)
        return self._rzusbstick_read_response()[0] == RZUSBStickResponses.RZ_RESP_SUCCESS

    # Discovery related functions
    def _get_capabilities(self):
        if "KILLERB" in self.__rzusbstick.product:
            capabilities = {
                Domain.Dot15d4 : (
                                    (Capability.Sniff | Capability.Inject),
                                    [Commands.Sniff, Commands.Send, Commands.Start, Commands.Stop]
                )
            }
        else:
            capabilities = {
                Domain.Dot15d4 : (
                                    (Capability.Sniff),
                                    [Commands.Sniff, Commands.Start, Commands.Stop]
                )
            }

        return capabilities

    def _get_manufacturer(self):
        if "KILLERB" in self.__rzusbstick.product:
            return "Joshua Wright (KillerBee version)".encode('utf-8')
        return (self.__rzusbstick.manufacturer + "(Factory version)").encode('utf-8')

    def _get_serial_number(self):
        return bytes.fromhex(
            self.__rzusbstick.serial_number[:-1] +
            f"{self.__rzusbstick.bus:04x}{self.__rzusbstick.address:04x}"
        )

    def _get_firmware_version(self):
        return (1, 0, 0)

    def _get_url(self):
        if "KILLERB" in self.__rzusbstick.product:
            return "https://github.com/riverloopsec/killerbee".encode('utf-8')

        return (
            "https://www.microchip.com/en-us/development-tool/ATAVRRZUSBSTICK"
        ).encode('utf-8')

    # RZUSBStick commands

    def _stop(self):
        stream_closed = self._close_stream()
        stopped = self._set_mode(RZUSBStickModes.RZ_MODE_NONE)
        return stream_closed and stopped

    def _start(self):
        started = self._set_mode(RZUSBStickModes.RZ_MODE_AIRCAPTURE)
        stream_enabled = self._open_stream()
        return started and stream_enabled

    def _set_mode(self, mode=RZUSBStickModes.RZ_MODE_NONE):
        return self._rzusbstick_send_command(RZUSBStickCommands.RZ_SET_MODE, data=[mode])

    def _set_channel(self, channel=11):
        if channel < 11 or channel > 26:
            return False
        self._close_stream()
        success = self._rzusbstick_send_command(RZUSBStickCommands.RZ_SET_CHANNEL, data=[channel])
        if success:
            self.__channel = channel
        self._open_stream()
        return success

    def _get_channel(self):
        return self.__channel

    def _enable_jamming(self):
        return self._rzusbstick_send_command(RZUSBStickCommands.RZ_JAMMER_ON)

    def _disable_jamming(self):
        return self._rzusbstick_send_command(RZUSBStickCommands.RZ_JAMMER_OFF)

    def _open_stream(self):
        if not self.__opened_stream:
            success = self._rzusbstick_send_command(RZUSBStickCommands.RZ_OPEN_STREAM)
            self.__opened_stream = success
        return self.__opened_stream

    def _close_stream(self):
        if self.__opened_stream:
            success = self._rzusbstick_send_command(RZUSBStickCommands.RZ_CLOSE_STREAM)
            self.__opened_stream = not success
            return True
        return False

    def _send_packet(self, data):
        if len(data) >= 1 and len(data) <= 125:
            self._close_stream()
            success = self._rzusbstick_send_command(RZUSBStickCommands.RZ_INJECT_FRAME,
                                                    bytes([len(data)])+data, timeout=700)
            self._open_stream()
            return success

        # Error.
        return False



"""
Exceptions related to RF4CE domain.
"""
class MissingRF4CEHeader(Exception):
    """RF4CE header is missing."""

    def __init__(self):
        """Constructor."""
        super().__init__()

class MissingRF4CESecurityFlag(Exception):
    """RF4CE Security Flag is missing.
    """
    def __init__(self):
        """Constructor."""
        super().__init__()

class MissingCryptographicMaterial(Exception):
    """Crypto material is missing."""

    def __init__(self):
        """Constructor."""
        super().__init__()
